"""Law module

This module contains classes that implement the main spectral distributions.
When the limiting behaviour of the spectrum of a random matrix ensemble is
well-known, it is often described with a mathematical proven law.
Probability density functions and cumulative distribution functions are provided
for the Wigner Semicircle Law, Marchenko-Pastur Law, Tracy-Widom Law and for the
spectrum of the Manova Ensemble.

"""

import numpy as np
import matplotlib.pyplot as plt
from scipy.integrate import quad
import collections.abc

from .gaussian_ensemble import GaussianEnsemble
from .wishart_ensemble import WishartEnsemble
from .manova_ensemble import ManovaEnsemble
from .tracy_widom_approximator import TW_Approximator


def _relu(x):
    """Element-wise maximum between the value and zero.

    Args:
        x (ndarray): list of numbers to compute its element-wise maximum.
    
    Returns:
        array_like consisting in the element-wise maximum vector of the given values.
    """
    return np.maximum(x, np.zeros_like(x))


def _indicator(x, start=None, stop=None, inclusive="both"):
    r"""Element-wise indicator function within a real interval.
    The interval can be left-closed, right-closed, closed or open.
    Visit https://en.wikipedia.org/wiki/Indicator_function for more information.

    Args:
        x (ndarray): list of numbers to compute its element-wise indicator image.
        start (double, default=None): left value of the interval. If not provided,
            the left value is equivalent to :math:`- \infty`.
        stop (double, default=None): right value of the interval. If not provided,
            the right value is equivalent to :math:`+ \infty`.
        inclusive (string, default="both"): type of interval. For left-closed interval
            use "left", for right-closed interval use "right", for closed interval use
            "both" and for open interval use "neither".

    Returns:
        array_like consisting in the element-wise indicator function image of the given values.
    """
    if start is None and stop is None:
        raise ValueError("Error: provide start and/or stop for indicator function.")

    INCLUSIVE_OPTIONS = ["both", "left", "right", "neither"]
    if inclusive not in INCLUSIVE_OPTIONS:
        raise ValueError(f"Error: invalid interval inclusive parameter: {inclusive}\n"
                         "\t inclusive has to be one of the following: {INCLUSIVE_OPTIONS}.")

    if start is not None:
        if inclusive == "both" or inclusive == "left":
            condition = (start <= x)
        elif inclusive == "neither" or inclusive == "right":
            condition = (start < x)
    
    if (start is not None) and (stop is not None):
        if inclusive == "both" or inclusive == "right":
            condition = np.logical_and(condition, (x <= stop))
        elif inclusive == "neither" or inclusive == "left":
            condition = np.logical_and(condition, (x < stop))
    elif stop:
        if inclusive == "both" or inclusive == "right":
            condition = (x <= stop)
        elif inclusive == "neither" or inclusive == "left":
            condition = (x < stop)

    return np.where(condition, 1.0, 0.0)


def _plot_func(interval, func, bins=1000, plot_title=None, plot_ylabel=None, savefig_path=None):
    """Plots a given function (callable) within the provided interval.

    Args:
        interval (tuple): Delimiters (xmin, xmax) of the histogram.
        func (callable): Function to be evaluated. The image of the function builds
            the y-axis values that are plotted.
        bins (int, default=100): It defines the number of equal-width bins within the
            provided interval or range.
        plot_title (string, default=None): Title of the plot.
        plot_ylabel (string, default=None): Label of the y-axis.
        savefig_path (string, default=None): path to save the created figure. If it is not
            provided, the plot is shown at the end of the routine.
    
    """
    if not isinstance(interval, tuple):
        raise ValueError("interval argument must be a tuple")
    
    (xmin, xmax) = interval

    xx = np.linspace(xmin, xmax, num=bins)
    yy = func(xx)

    plt.plot(xx, yy)
    plt.xlabel("x")
    if plot_ylabel:
        plt.ylabel(plot_ylabel)
    
    if plot_title:
        plt.title(plot_title)

    if savefig_path:
        plt.savefig(savefig_path, dpi=800)
    else:
        plt.show()


def _get_bins_centers_and_contour(bins):
    """Calculates the centers and contour of the given bins.

    Computes the centers of the given bins. Also, the smallest and the largest bin
    delimitiers are included to define the countour of the representation interval.

    Args:
        bins (list): list of numbers (floats) that specify each bin delimiter.

    Returns:
        list of numbers (floats) consisting in the list of bin centers and contour.
    
    """
    centers = [bins[0]] # Adding initial contour
    l = len(bins)
    for i in range(l-1):
        centers.append((bins[i]+bins[i+1])/2) # Adding centers
    centers.append(bins[-1]) # Adding final contour
    return centers


class WignerSemicircleDistribution:
    """Wigner Semicircle Distribution class.

    The Wigner Semicircle Law describes the spectrum of the Wigner random matrices.
    In particular, random matrices of the Gaussian Ensemble are Wigner matrices,
    and therefore the spectrum of the random matrices of this ensemble converge
    to the Wigner Semicircle Law when the matrix size goes to infinity. This
    class provides methods to sample eigenvalues following Wigner Semicircle
    distribution, computing the PDF, computing the CDF and simple methods to
    plot the former two.

    Attributes:
        beta (int): descriptive integer of the Gaussian ensemble type.
            For GOE beta=1, for GUE beta=2, for GSE beta=4.
        center (float): center of the distribution. Since the distribution
            has the shape of a semicircle, the center corresponds to its center.
        sigma (float): scale of the distribution. This value also corresponds
            to the standard deviation of the random entries of the sampled matrix.
        radius (float): radius of the semicircle of the Wigner law. This depends on
            the scale (sigma) and on beta.
    
    References:
        - Wigner, E.
            "Characteristic Vectors of Bordered Matrices With Infinite Dimensions".
            Annals of Mathematics. 62.3. (1955).
        - Wigner, E.
            "On the Distribution of the Roots of Certain Symmetric Matrices".
            Annals of Mathematics. 67.2. (1958).
    
    """

    def __init__(self, beta=1, center=0.0, sigma=1.0):
        """Constructor for WignerSemicircleDistribution class.

        Initializes an instance of this class with the given parameters.

        Args:
            beta (int, default=1): descriptive integer of the Gaussian ensemble type.
                For GOE beta=1, for GUE beta=2, for GSE beta=4.
            center (float, default=0.0): center of the distribution. Since the distribution
                has the shape of a semicircle, the center corresponds to its center.
            sigma (float, default=1.0): scale of the distribution. This value also corresponds
                to the standard deviation of the random entries of the sampled matrix.
        
        """
        if beta not in [1,2,4]:
            raise ValueError(f"Error: invalid beta. It has to be 1,2 or 4. Provided beta = {beta}.")

        self.beta = beta
        self.center = center
        self.sigma = sigma
        self.radius = 2.0 * np.sqrt(self.beta) * self.sigma
    
    def rvs(self, size):
        """Samples ranfom variates following this distribution.
        This uses the relationship between Wigner Semicircle law and Beta distribution.

        Args:
            size (int): sample size.
        
        Returns:
            numpy array with the generated samples.
        """
        if size <= 0:
            raise ValueError(f"Error: invalid sample size. It has to be positive. Provided size = {size}.")

        # Use relationship with beta distribution
        beta_samples = np.random.beta(1.5, 1.5, size=size)
        return self.center + 2*self.radius*beta_samples - self.radius

    def pdf(self, x):
        """Computes PDF of the Wigner Semicircle Law.

        Args:
            x (float or ndarray): value or (numpy) array of values in which compute the PDF.
        
        Returns:
            float or numpy array with the computed PDF in the given value(s).
        
        """
        return 2.0 * np.sqrt(_relu(self.radius**2 - (x-self.center)**2)) / (np.pi * self.radius**2)
    
    def cdf(self, x):
        """Computes CDF of the Wigner Semicircle Law.

        Args:
            x (float or ndarray): value or (numpy) array of values in which compute the CDF.
        
        Returns:
            float or numpy array with the computed CDF in the given value(s).
        
        """
        with np.errstate(divide='ignore', invalid='ignore'):
            return np.select(
                condlist=[x >= (self.center + self.radius), x <= (self.center - self.radius)],
                choicelist=[1.0, 0.0],
                default=(0.5 + ((x-self.center) * np.sqrt(self.radius**2 - (x-self.center)**2))/(np.pi * self.radius**2) \
                         + (np.arcsin((x-self.center)/self.radius)) / np.pi)
            )
    
    def plot_pdf(self, interval=None, bins=1000, savefig_path=None):
        """Plots the PDF of the Wigner Semicircle Law.

        Args:
            interval (tuple, default=None): Delimiters (xmin, xmax) of the histogram. If not
                provided, the used interval is calculated depending on beta, center, radius and scale.
            bins (int, default=100): It defines the number of equal-width bins within the
                provided interval or range.
            savefig_path (string, default=None): path to save the created figure. If it is not
                provided, the plot is shown at the end of the routine.
        
        """
        if interval is None:
            interval = (self.center - self.radius - 0.1, self.center + self.radius + 0.1)
        
        _plot_func(
            interval, func=self.pdf, bins=bins, plot_title="Wigner Semicircle law PDF", 
            plot_ylabel="probability density", savefig_path=savefig_path
        )
    
    def plot_cdf(self, interval=None, bins=1000, savefig_path=None):
        """Plots the CDF of the Wigner Semicircle Law.

        Args:
            interval (tuple, default=None): Delimiters (xmin, xmax) of the histogram. If not
                provided, the used interval is calculated depending on beta, center, radius and scale.
            bins (int, default=100): It defines the number of equal-width bins within the
                provided interval or range.
            savefig_path (string, default=None): path to save the created figure. If it is not
                provided, the plot is shown at the end of the routine.
        
        """
        if interval is None:
            interval = (self.center - self.radius - 0.1, self.center + self.radius + 0.1)
        
        _plot_func(
            interval, func=self.cdf, bins=bins, plot_title="Wigner Semicircle law CDF",
            plot_ylabel="cumulative distribution", savefig_path=savefig_path
        )
    

    def plot_empirical_pdf(self, n_size=10000, bins=100, interval=None, density=False,
                           plot_law_pdf=False, savefig_path=None):
        """Computes and plots Wigner's semicircle empirical law.

        Calculates and plots Wigner's semicircle empirical law using random samples generated
        using the relationship between the Wigner Semicircle law and the Beta distribution:
        the Wigner's Semicircle distribution it is a scaled Beta distribution with parameters
        :math:`\alpha = \beta = 3/2`.

        Args:
            n_size (int, default=1000): number of random samples that can be interpreted as
                random eigenvalues of a Wigner matrix. This is the sample size.
            bins (int or sequence, default=100): If bins is an integer, it defines the number
                of equal-width bins in the range. If bins is a sequence, it defines the
                bin edges, including the left edge of the first bin and the right
                edge of the last bin; in this case, bins may be unequally spaced.
            interval (tuple, default=None): Delimiters (xmin, xmax) of the histogram.
                The lower and upper range of the bins. Lower and upper outliers are ignored.
            density (bool, default=False): If True, draw and return a probability
                density: each bin will display the bin's raw count divided by the total
                number of counts and the bin width, so that the area under the histogram
                integrates to 1. If set to False, the absolute frequencies of the eigenvalues
                are returned.
            plot_law_pdf (bool, default=False): If True, the theoretical law is plotted.
                If set to False, just the empirical histogram is shown. This parameter is only
                considered when the argument 'density' is set also to True.
            savefig_path (string, default=None): path to save the created figure. If it is not
                provided, the plot is shown are the end of the routine.

        References:
            - Albrecht, J. and Chan, C.P. and Edelman, A.
                "Sturm sequences and random eigenvalue distributions".
                Foundations of Computational Mathematics. 9.4 (2008): 461-483.
            - Dumitriu, I. and Edelman, A.
                "Matrix Models for Beta Ensembles".
                Journal of Mathematical Physics. 43.11 (2002): 5830-5847.

        """
        # pylint: disable=too-many-arguments
        if n_size<1:
            raise ValueError("matrix size must be positive")
        
        if interval is None:
            interval = (-self.radius, self.radius)

        random_samples = self.rvs(size=n_size)
        observed, bins = np.histogram(random_samples, bins=bins, range=interval, density=density)

        width = bins[1]-bins[0]
        plt.bar(bins[:-1], observed, width=width, align='edge')

        # Plotting Wigner Semicircle Law pdf
        if plot_law_pdf and density:
            centers = np.asarray(_get_bins_centers_and_contour(bins))
            pdf = self.pdf(centers)
            plt.plot(centers, pdf, color='red', linewidth=2)
        elif plot_law_pdf and not density:
            print("Warning: Wigner's Semicircle Law PDF is only plotted when density is True.")

        plt.title("Wigner Semicircle Law - Eigenvalue histogram", fontweight="bold")
        plt.xlabel("x")
        plt.ylabel("probability density")

        # Saving plot or showing it
        if savefig_path:
            plt.savefig(savefig_path, dpi=1200)
        else:
            plt.show()



class MarchenkoPasturDistribution:
    """Marchenko-Pastur Distribution class.

    The Marchenko-Pastur Law describes the spectrum of the Wishart random matrices.
    Therefore the spectrum of the random matrices of this ensemble converge
    to the Marchenko-Pastur Law when the matrix size goes to infinity. This
    class provides methods to sample eigenvalues following Marchenko-Pastur
    distribution, computing the PDF, computing the CDF and simple methods to
    plot the former two.

    Attributes:
        ratio (float): random matrix size ratio. This is the ratio between the
            number of degrees of freedom 'p' and the sample size 'n'. The value
            of ratio = p/n.
        beta (int): descriptive integer of the Wishart ensemble type.
            For WRE beta=1, for WCEE beta=2, for WQE beta=4.
        sigma (float): scale of the distribution. This value also corresponds
            to the standard deviation of the random entries of the sampled matrix.
        lambda_minus (float): lower bound of the support of the Marchenko-Pastur Law.
            It depends on beta, on the scale (sigma) and on the ratio.
        lambda_plus (float): upper bound of the support of the Marchenko-Pastur Law.
            It depends on beta, on the scale (sigma) and on the ratio.
    
    References:
        - Bar, Z.D. and Silverstain, J.W.
            "Spectral Analysis of Large Dimensional Random Matrices".
            2nd edition. Springer. (2010).
    
    """

    ARCTAN_OF_INFTY = np.pi/2

    def __init__(self, ratio, beta=1, sigma=1.0):
        """Constructor for MarchenkoPasturDistribution class.

        Initializes an instance of this class with the given parameters.

        Args:
            ratio (float): random matrix size ratio (:math:`\lambda`). This is the ratio
                between the number of degrees of freedom :math:`p` and the sample size :math:`n`.
                The value of ratio is computed as :math:`\lambda = p/n`.
            beta (int, default=1): descriptive integer of the Wishart ensemble type (:math:`\beta`).
                For WRE beta=1, for WCE beta=2, for WQE beta=4.
            sigma (float, default=1.0): scale of the distribution (:math:`\sigma`). This value also
                corresponds to the standard deviation of the random entries of the sampled matrix.
        
        """
        if beta not in [1,2,4]:
            raise ValueError(f"Error: invalid beta. It has to be 1,2 or 4. Provided beta = {beta}.")
        if ratio <= 0:
            raise ValueError(f"Error: invalid ratio. It has to be positive. Provided ratio = {ratio}.")

        self.ratio = ratio
        self.beta = beta
        self.sigma = sigma
        self.lambda_minus = self.beta * self.sigma**2 * (1 - np.sqrt(self.ratio))**2
        self.lambda_plus = self.beta * self.sigma**2 * (1 + np.sqrt(self.ratio))**2
        self._var = self.beta * self.sigma**2
        self._wishart_ens = None
    
    def rvs(self, size):
        """Samples ranfom variates following this distribution.

        Args:
            size (int): sample size :math:`n`.
        
        Returns:
            numpy array with the generated samples.
        """
        if size <= 0:
            raise ValueError(f"Error: invalid sample size. It has to be positive. Provided size = {size}.")
        
        _n = int(np.round(size / self.ratio))

        if not self._wishart_ens:
            self._wishart_ens = WishartEnsemble(beta=self.beta, p=size, n=_n, use_tridiagonal=False, sigma=self.sigma)
        else:
            self._wishart_ens.set_size(p=size, n=_n, resample_mtx=True)

        if self.beta == 4:
            return self._wishart_ens.eigvals(normalize=True)[::2]
        return self._wishart_ens.eigvals(normalize=True)

    def pdf(self, x):
        """Computes PDF of the Marchenko-Pastur Law.

        Args:
            x (float or ndarray): value or (numpy) array of values in which compute the PDF.
        
        Returns:
            float or numpy array with the computed PDF in the given value(s).
        
        """
        with np.errstate(divide='ignore', invalid='ignore'):
            return np.sqrt(_relu(self.lambda_plus - x) * _relu(x - self.lambda_minus)) \
                / (2.0 * np.pi * self.ratio * self._var * x)

    def cdf(self, x):
        """Computes CDF of the Marchenko-Pastur Law.

        Args:
            x (float or ndarray): value or (numpy) array of values in which compute the CDF.
        
        Returns:
            float or numpy array with the computed CDF in the given value(s).
        
        """
        with np.errstate(divide='ignore', invalid='ignore'):
            acum = _indicator(x, start=self.lambda_plus, inclusive="left")
            acum += np.where(_indicator(x, start=self.lambda_minus, stop=self.lambda_plus, inclusive="left"),
                            self._cdf_aux_f(x), 0.0)

            if self.ratio <= 1:
                return acum
            
            acum += np.where(_indicator(x, start=self.lambda_minus, stop=self.lambda_plus, inclusive="left"),
                            (self.ratio-1)/(2*self.ratio), 0.0)

            ### This would need to be added if the extra density point at zero is measured
            # https://en.wikipedia.org/wiki/Marchenko%E2%80%93Pastur_distribution
            # acum += np.where(_indicator(x, start=0, stop=self.lambda_minus, inclusive="left"),
            #                 (self.ratio-1)/self.ratio, 0.0)
            return acum

    def _cdf_aux_f(self, x):
        first_arctan_term = np.where(x == self.lambda_minus,
                                    MarchenkoPasturDistribution.ARCTAN_OF_INFTY,
                                    np.arctan((self._cdf_aux_r(x)**2 - 1)/(2 * self._cdf_aux_r(x)))
                                )

        second_arctan_term = np.where(x == self.lambda_minus,
                                      MarchenkoPasturDistribution.ARCTAN_OF_INFTY,
                                      np.arctan((self.lambda_minus*self._cdf_aux_r(x)**2 - self.lambda_plus) \
                                                / (2*self._var*(1-self.ratio)*self._cdf_aux_r(x)))
                                )
        return 1/(2*np.pi*self.ratio) * (np.pi*self.ratio \
                                         + (1/self._var)*np.sqrt(_relu(self.lambda_plus-x)*_relu(x-self.lambda_minus)) \
                                         - (1+self.ratio)*first_arctan_term + (1-self.ratio)*second_arctan_term)

    def _cdf_aux_r(self, x):
        return np.sqrt((self.lambda_plus-x)/(x - self.lambda_minus))

    def plot_pdf(self, interval=None, bins=1000, savefig_path=None):
        """Plots the PDF of the Marchenko-Pastur Law.

        Args:
            interval (tuple, default=None): Delimiters (xmin, xmax) of the histogram. If not
                provided, the used interval is calculated depending on beta, ratio, and scale.
            bins (int, default=100): It defines the number of equal-width bins within the
                provided interval or range.
            savefig_path (string, default=None): path to save the created figure. If it is not
                provided, the plot is shown at the end of the routine.
        
        """
        if interval is None:
            interval = (self.lambda_minus, self.lambda_plus)
        
        _plot_func(
            interval, func=self.pdf, bins=bins, plot_title="Marchenko-Pastur law PDF",
            plot_ylabel="probability density", savefig_path=savefig_path
        )
    
    def plot_cdf(self, interval=None, bins=1000, savefig_path=None):
        """Plots the CDF of the Marchenko-Pastur Law.

        Args:
            interval (tuple, default=None): Delimiters (xmin, xmax) of the histogram. If not
                provided, the used interval is calculated depending on beta, ratio, and scale.
            bins (int, default=100): It defines the number of equal-width bins within the
                provided interval or range.
            savefig_path (string, default=None): path to save the created figure. If it is not
                provided, the plot is shown at the end of the routine.
        
        """
        if interval is None:
            interval = (self.lambda_minus, self.lambda_plus)
        
        _plot_func(
            interval, func=self.cdf, bins=bins, plot_title="Marchenko-Pastur law CDF",
            plot_ylabel="cumulative distribution", savefig_path=savefig_path
        )

    def plot_empirical_pdf(self, n_size=3000, p_size=None, bins=100, interval=None,
                           density=False, plot_law_pdf=False, savefig_path=None):
        """Computes and plots Marchenko-Pastur empirical law using Wishart Ensemble random matrices.

        Calculates and plots Marchenko-Pastur empirical law using Wishart Ensemble random matrices.
        The size of the sampled matrix will depend on the `n_size` (:math:`n`) parameter and on the
        ratio :math:`\lambda` given when instantiating this class, unless the parameter `p_size` (:math:`p`)
        is also given. In this last case, the ratio :math:`\lambda` for the empirical pdf plotting is
        computed as :math:`\lambda = p/n`. If only the sample size :math:`n` is provided, the number
        of degrees of freedom :math:`p` is computed as :math:`[\lambda * n]`.
        Wishart (Laguerre) ensemble has improved routines (using tridiagonal forms and Sturm
        sequences) to avoid calculating the eigenvalues, so the histogram is built using certain
        techniques to boost efficiency. This optimization is only used when the ratio p_size/n_size
        is less or equal than 1.

        Args:
            n_size (int, default=3000): number of columns of the guassian matrix that generates
                the matrix of the corresponding ensemble. This is the sample size. The number of
                degrees of freedom is computed depending on this argument and on the given ratio,
                unless the argument `p_size` is also provided, which in this case the ratio is
                re-computed as ratio=p_size/n_size.
            p_size (int, default=None): number of rows of the guassian matrix that generates
                the matrix of the corresponding ensemble. If provided, the current ratio is ignored
                (but not replaced) and the new ratio=p_size/n_size is used instead.
            bins (int or sequence, default=100): If bins is an integer, it defines the number
                of equal-width bins in the range. If bins is a sequence, it defines the
                bin edges, including the left edge of the first bin and the right
                edge of the last bin; in this case, bins may be unequally spaced.
            interval (tuple, default=None): Delimiters (xmin, xmax) of the histogram.
                The lower and upper range of the bins. Lower and upper outliers are ignored.
            density (bool, default=False): If True, draw and return a probability
                density: each bin will display the bin's raw count divided by the total
                number of counts and the bin width, so that the area under the histogram
                integrates to 1. If set to False, the absolute frequencies of the eigenvalues
                are returned.
            plot_law_pdf (bool, default=False): If True, the limiting theoretical law is plotted.
                If set to False, just the empirical histogram is shown. This parameter is only
                considered when the argument 'density' is set also to True.
            savefig_path (string, default=None): path to save the created figure. If it is not
                provided, the plot is shown are the end of the routine.

        References:
            - Albrecht, J. and Chan, C.P. and Edelman, A.
                "Sturm sequences and random eigenvalue distributions".
                Foundations of Computational Mathematics. 9.4 (2008): 461-483.
            - Dumitriu, I. and Edelman, A.
                "Matrix Models for Beta Ensembles".
                Journal of Mathematical Physics. 43.11 (2002): 5830-5847.

        """
        # pylint: disable=too-many-arguments
        if n_size<1 or (p_size is not None and p_size<1):
            raise ValueError("matrix size must be positive")
        
        if interval and interval[0] == 0:
            print("Warning: setting the beginning of the interval to zero may generate numerical errors.")
            print(f"Setting interval to (-0.01, {interval[1]})")
            interval = (-0.01, interval[1])

        # calculating constants depending on matrix sizes
        if p_size is None:
            p_size = round(self.ratio * n_size)

        # computing an approximated ratio since p_size is rounded to the closest integer        
        approx_ratio = p_size/n_size
        lambda_plus = self.beta * self.sigma**2 * (1 + np.sqrt(approx_ratio))**2
        lambda_minus = self.beta * self.sigma**2 * (1 - np.sqrt(approx_ratio))**2
        use_tridiag_ratio = (approx_ratio <= 1)
        if not use_tridiag_ratio:
            print("Warning: Cannot use tridiagonal histogramming if 'p' (degrees of freedom) is "
                " greater than 'n' (sample size).\n"
                f"\t Provided n={n_size} and p={p_size}. Tridiagonal histogramming is therefore deactivated.\n"
                "\t It is adviced to increase sample size (`n`) to optimize and boost histogramming.")

        # computing interval according to the matrix size ratio and support
        if interval is None:
            if approx_ratio <= 1:
                interval = (lambda_minus, lambda_plus)
            else:
                interval = (min(-0.05, lambda_minus), lambda_plus)
        
        use_tridiag_sigma = (self.sigma == 1.0)
        if not use_tridiag_sigma:
            print(f"Warning: The given scale is not the standard (sigma = {self.sigma}).\n"
                "\t Tridiagonal histogramming is deactivated.\n"
                "\t It is adviced to set sigma=1.0 to optimize and boost histogramming.")

        ens = WishartEnsemble(beta=self.beta, p=p_size, n=n_size, sigma=self.sigma,
                            use_tridiagonal=(use_tridiag_ratio and use_tridiag_sigma))

        observed, bins = ens.eigval_hist(bins=bins, interval=interval, density=density, normalize=True)

        width = bins[1]-bins[0]
        plt.bar(bins[:-1], observed, width=width, align='edge')

        # Plotting theoretical graphic
        if plot_law_pdf and density:
            centers = np.array(_get_bins_centers_and_contour(bins))
            # creating new instance with the approximated ratio depending on the given matrix sizes
            mpd = MarchenkoPasturDistribution(beta=self.beta, ratio=approx_ratio, sigma=self.sigma)
            pdf = mpd.pdf(centers)
            plt.plot(centers, pdf, color='red', linewidth=2)

        plt.title("Marchenko-Pastur Law - Eigenvalue histogram", fontweight="bold")
        plt.xlabel("x")
        plt.ylabel("probability density")
        if approx_ratio > 1:
            if plot_law_pdf and density:
                ylim_vals = pdf
            else:
                ylim_vals = observed
            try:
                plt.ylim(0, np.max(ylim_vals)+0.25*np.max(ylim_vals))
            except ValueError:
                second_highest_val = np.partition(ylim_vals.flatten(), -2)[-2]
                plt.ylim(0, second_highest_val+0.25*second_highest_val)

        # Saving plot or showing it
        if savefig_path:
            plt.savefig(savefig_path, dpi=1200)
        else:
            plt.show()



class TracyWidomDistribution:
    """Tracy-Widom Distribution class.

    The Tracy-Widom Law describes the behaviour of the largest eigenvalue of the
    Wigner random matrices. In particular, random matrices of the Gaussian Ensemble
    are Wigner matrices, and therefore the largest eigenvalues of the spectrum of
    the random matrices of this ensemble converge to the Tracy-Widom Law when the
    matrix size and the sample size (number of times a Wigner matrix is generated to
    compute the largest eigenvalue) goes to infinity. This class provides methods to
    sample eigenvalues following Tracy-Widom distribution, computing the PDF,
    computing the CDF and simple methods to plot the former two.

    Attributes:
        beta (int): descriptive integer of the Gaussian ensemble type.
            For GOE beta=1, for GUE beta=2, for GSE beta=4.
    
    References:
        - Bauman, S.
            "The Tracy-Widom Distribution and its Application to Statistical Physics".
            http://web.mit.edu/8.334/www/grades/projects/projects17/SamBauman.pdf
            MIT Department of Physics. (2017).
        - Tracy, C.A. and Widom, H.
            "On orthogonal and symplectic matrix ensembles".
            Communications in Mathematical Physics. 177.3. (1996).
        - Bejan, A.
            "Largest eigenvalues and sample covariance matrices".
            Tracy-Widom and Painleve II: Computational aspects and
            realization in S-Plus with applications, M.Sc. dissertation,
            Department of Statistics, The University of Warwick. (2005).
        - Borot, G. and Nadal, C.
            "Right tail expansion of Tracy-Widom beta laws".
            Random Matrices: Theory and Applications. 01.03. (2012).
    
    """

    def __init__(self, beta=1):
        """Constructor for TracyWidomDistribution class.

        Initializes an instance of this class with the given parameters.

        Args:
            beta (int, default=1): descriptive integer of the Gaussian ensemble type.
                For GOE beta=1, for GUE beta=2, for GSE beta=4.
        
        """
        if beta not in [1,2,4]:
            raise ValueError(f"Error: invalid beta. It has to be 1,2 or 4. Provided beta = {beta}.")

        self.beta = beta
        self.tw_approx = TW_Approximator(beta=self.beta)

    def rvs(self, size, mtx_size=100):
        """Samples ranfom variates following this distribution.

        Args:
            size (int): sample size.
            mtx_size (int, default=100): matrix size. Remember the Tracy-Widom Law describes the
                limiting behaviour of the largest eigenvalue of a Wigner matrix. Therefore,
                a matrix has to be generated to get each sample. This argument specifies the size
                of the matrix.
        
        Returns:
            numpy array with the generated samples.
        """
        if size <= 0:
            raise ValueError(f"Error: invalid sample size. It has to be positive. Provided size = {size}.")
        if mtx_size <= 0:
            raise ValueError(f"Error: invalid matrix size. It has to be positive. Provided matrix size = {mtx_size}.")

        self._gaussian_ens = GaussianEnsemble(beta=self.beta, n=mtx_size, use_tridiagonal=False)

        max_eigvals = []
        for _ in range(size):
            max_eigvals.append(self._gaussian_ens.eigvals(normalize=False).max())
            self._gaussian_ens.sample()
        max_eigvals = np.asarray(max_eigvals)

        # Tracy-Widom eigenvalue distr. normalization constants
        eigval_scale = 1.0
        size_scale = 1.0
        if self.beta == 2:
            eigval_scale = 1/np.sqrt(2)
        if self.beta == 4:
            eigval_scale = size_scale = 1/np.sqrt(2)
            mtx_size *= 2

        max_eigvals = size_scale*(mtx_size**(1/6))*(eigval_scale*max_eigvals - (2.0*np.sqrt(mtx_size)))
        return max_eigvals

    def pdf(self, x):
        """Computes PDF of the Tracy-Widom Law.

        Args:
            x (float or ndarray): value or (numpy) array of values in which compute the PDF.
        
        Returns:
            float or numpy array with the computed PDF in the given value(s).
        
        """
        return self.tw_approx.pdf(x)

    def cdf(self, x):
        """Computes CDF of the Tracy-Widom Law.

        Args:
            x (float or ndarray): value or (numpy) array of values in which compute the CDF.
        
        Returns:
            float or numpy array with the computed CDF in the given value(s).
        
        """
        return self.tw_approx.cdf(x)

    def plot_pdf(self, interval=None, bins=1000, savefig_path=None):
        """Plots the PDF of the Tracy-Widom Law.

        Args:
            interval (tuple, default=None): Delimiters (xmin, xmax) of the histogram. If not
                provided, the used interval is calculated depending on beta.
            bins (int, default=100): It defines the number of equal-width bins within the
                provided interval or range.
            savefig_path (string, default=None): path to save the created figure. If it is not
                provided, the plot is shown at the end of the routine.
        
        """
        if interval is None:
            interval = (-5, 4-self.beta)
        
        _plot_func(
            interval, func=self.pdf, bins=bins, plot_title="Tracy-Widom law PDF",
            plot_ylabel="probability density", savefig_path=savefig_path
        )
    
    def plot_cdf(self, interval=None, bins=1000, savefig_path=None):
        """Plots the PDF of the Tracy-Widom Law.

        Args:
            interval (tuple, default=None): Delimiters (xmin, xmax) of the histogram. If not
                provided, the used interval is calculated depending on beta.
            bins (int, default=100): It defines the number of equal-width bins within the
                provided interval or range.
            savefig_path (string, default=None): path to save the created figure. If it is not
                provided, the plot is shown at the end of the routine.
        
        """
        if interval is None:
            interval = (-5, 4-self.beta)
        
        _plot_func(
            interval, func=self.cdf, bins=bins, plot_title="Tracy-Widom law CDF",
            plot_ylabel="cumulative distribution", savefig_path=savefig_path
        )

    def plot_empirical_pdf(self, n_size=100, times=1000, bins=100, interval=None,
                           density=False, plot_law_pdf=False, savefig_path=None):
        """Computes and plots Tracy-Widom empirical law using Gaussian Ensemble.

        Calculates and plots Tracy-Widom empirical law using Gaussian Ensemble random matrices.
        Because we need to obtain the largest eigenvalue of each sampled random matrix,
        we need to sample a certain amount them. For each random matrix sammpled, its
        largest eigenvalue is calcualted in order to simulate its density.

        Args:
            n_size (int, default=100): random matrix size n times n. This is the sample size.
            times (int, default=1000): number of times to sample a random matrix.
            bins (int or sequence, default=100): If bins is an integer, it defines the number
                of equal-width bins in the range. If bins is a sequence, it defines the
                bin edges, including the left edge of the first bin and the right
                edge of the last bin; in this case, bins may be unequally spaced.
            interval (tuple, default=None): Delimiters (xmin, xmax) of the histogram.
                The lower and upper range of the bins. Lower and upper outliers are ignored.
            density (bool, default=False): If True, draw and return a probability
                density: each bin will display the bin's raw count divided by the total
                number of counts and the bin width, so that the area under the histogram
                integrates to 1. If set to False, the absolute frequencies of the eigenvalues
                are returned.
            plot_law_pdf (bool, default=False): If True, the limiting theoretical law is plotted.
                If set to False, just the empirical histogram is shown. This parameter is only
                considered when the argument 'density' is set also to True.
            savefig_path (string, default=None): path to save the created figure. If it is not
                provided, the plot is shown are the end of the routine.

        References:
            - Albrecht, J. and Chan, C.P. and Edelman, A.
                "Sturm sequences and random eigenvalue distributions".
                Foundations of Computational Mathematics. 9.4 (2008): 461-483.
            - Dumitriu, I. and Edelman, A.
                "Matrix Models for Beta Ensembles".
                Journal of Mathematical Physics. 43.11 (2002): 5830-5847.

        """
        # pylint: disable=too-many-arguments
        if n_size<1 or times<1:
            raise ValueError("matrix size or number of repetitions must be positive")

        ens = GaussianEnsemble(beta=self.beta, n=n_size, use_tridiagonal=False)

        eigvals = []
        for _ in range(times):
            eigvals.append(ens.eigvals(normalize=False).max())
            ens.sample()
        eigvals = np.asarray(eigvals)

        # Tracy-Widom eigenvalue distr. normalization constants
        eigval_scale = 1.0
        size_scale = 1.0
        if self.beta == 2:
            eigval_scale = 1/np.sqrt(2)
        elif self.beta == 4:
            eigval_scale = size_scale = 1/np.sqrt(2)
            n_size *= 2
        eigvals = size_scale*(n_size**(1/6))*(eigval_scale*eigvals - (2.0*np.sqrt(n_size)))

        if interval is None:
            xmin=eigvals.min()
            xmax=eigvals.max()
            interval=(xmin, xmax)

        # using numpy to obtain histogram in the given interval and no. of bins
        observed, bins = np.histogram(eigvals, bins=bins, range=interval, density=density)
        width = bins[1]-bins[0]
        plt.bar(bins[:-1], observed, width=width, align='edge')

        # Plotting theoretical graphic
        if plot_law_pdf and density:
            centers = _get_bins_centers_and_contour(bins)
            pdf = self.pdf(centers)
            plt.plot(centers, pdf, color='red', linewidth=2)

        plt.title("Tracy-Widom Law - Eigenvalue histogram", fontweight="bold")
        plt.xlabel("x")
        plt.ylabel("probability density")

        # Saving plot or showing it
        if savefig_path:
            plt.savefig(savefig_path, dpi=1200)
        else:
            plt.show()



class ManovaSpectrumDistribution:
    """Manova Spectrum Distribution class.

    The spectrum of the random matrices of the Manova Ensemble converge to a
    well-defined function implemented in this class. The class provides methods
    to sample eigenvalues of the Manova Ensemble, computing the PDF, computing
    the CDF and simple methods to plot the former two.

    Attributes:
        a (float): first random matrix size ratio. This is the ratio between the
            number of degrees of freedom 'p' and the first sample size 'n1'. The value
            of a = p/n1. Remember a Manova randon matrix is considered a double-Wishart
            matrix, that's why there are two sample sizes 'n1' and 'n2' (see below).
        b (float): second random matrix size ratio. This is the ratio between the
            number of degrees of freedom 'p' and the second sample size 'n2'. The value
            of b = p/n2. Remember a Manova randon matrix is considered a double-Wishart
            matrix, that's why there are two sample sizes 'n1' and 'n2'.
        beta (int, default=1): descriptive integer of the Manova ensemble type.
            For MRE beta=1, for WME beta=2, for MQE beta=4.
        sigma (float): scale of the distribution. This value also corresponds
            to the standard deviation of the random entries of the sampled matrix.
        lambda_minus (float): lower bound of the support of the Manova spectrum distribution.
            It depends on beta, on the scale (sigma) and on the ratio.
        lambda_plus (float): upper bound of the support of the Manova spectrum distribution.
            It depends on beta, on the scale (sigma) and on the ratio.
    
    References:
        - Erdos, L. and Farrell, B.
            "Local Eigenvalue Density for General MANOVA Matrices".
            Journal of Statistical Physics. 152.6 (2013): 1003-1032.
    
    """

    def __init__(self, a, b, beta=1):
        """Constructor for ManovaSpectrumDistribution class.

        Initializes an instance of this class with the given parameters.

        Args:
            a (float): first random matrix size ratio. This is the ratio between the
                number of degrees of freedom 'p' and the first sample size 'n1'. The value
                of a = p/n1. Remember a Manova randon matrix is considered a double-Wishart
                matrix, that's why there are two sample sizes 'n1' and 'n2' (see below).
            b (float): second random matrix size ratio. This is the ratio between the
                number of degrees of freedom 'p' and the second sample size 'n2'. The value
                of b = p/n2. Remember a Manova randon matrix is considered a double-Wishart
                matrix, that's why there are two sample sizes 'n1' and 'n2'.
            beta (int, default=1): descriptive integer of the Manova ensemble type.
                For MRE beta=1, for WME beta=2, for MQE beta=4.
        
        """
        if beta not in [1,2,4]:
            raise ValueError(f"Error: invalid beta. It has to be 1,2 or 4. Provided beta = {beta}.")
        if a <= 0 or b <= 0:
            raise ValueError("Error: invalid matrix parameters. They have to be both positive.\n"
                             f"\tProvided a = {a} and b = {b}.")

        if a < 1 or b < 1:
            print(f"Warning: Setting a < 1 (a = {a}) or b < 1 (b = {b}) may cause numerical instability.")

        self.a = a
        self.b = b
        self.beta = beta
        self.lambda_term1 = np.sqrt((a/(a+b)) * (1 - (1/(a+b))))
        self.lambda_term2 = np.sqrt((1/(a+b)) * (1 - (a/(a+b))))
        self.lambda_minus = (self.lambda_term1 - self.lambda_term2)**2
        self.lambda_plus = (self.lambda_term1 + self.lambda_term2)**2
        self._manova_ens = None
    
    def rvs(self, size):
        """Samples random variates following this distribution.

        Args:
            size (int): sample size.
        
        Returns:
            numpy array with the generated samples.
        """
        if size <= 0:
            raise ValueError(f"Error: invalid sample size. It has to be positive. Provided size = {size}.")
        
        _n1 = int(np.round(size * self.a))
        _n2 = int(np.round(size * self.b))

        if not self._manova_ens:
            self._manova_ens = ManovaEnsemble(beta=self.beta, m=size, n1=_n1, n2=_n2)
        else:
            self._manova_ens.set_size(m=size, n1=_n1, n2=_n2, resample_mtx=True)

        # normalization here is not crucial since the default normalization const. of Manova is 1.0
        if self.beta == 4:
            return self._manova_ens.eigvals(normalize=True)[::2].real
        return self._manova_ens.eigvals(normalize=True).real
    
    def pdf(self, x):
        """Computes PDF of the Manova Spectrum distribution.

        Args:
            x (float or ndarray): value or (numpy) array of values in which compute the PDF.
        
        Returns:
            float or numpy array with the computed PDF in the given value(s).
        
        """
        with np.errstate(divide='ignore', invalid='ignore'): 
            return np.where(np.logical_and(x > self.lambda_minus, x < self.lambda_plus),
                            (self.a + self.b) * np.sqrt((self.lambda_plus - x) * (x - self.lambda_minus)) \
                                / (2.0 * np.pi * x * (1-x)),
                            0.0)
    
    def __cdf(self, x):
        if x <= self.lambda_minus:
            return 0.0

        if x >= self.lambda_plus:
            return 1.0

        return quad(self.pdf, self.lambda_minus, x)[0]

    def cdf(self, x):
        """Computes CDF of the Manova Spectrum distribution.

        Args:
            x (float or ndarray): value or (numpy) array of values in which compute the CDF.
        
        Returns:
            float or numpy array with the computed CDF in the given value(s).
        
        """
        # if x is array-like
        if isinstance(x, (collections.abc.Sequence, np.ndarray)):
            y_ret = []
            for val in x:
                y_ret.append(self.__cdf(val))
            return np.asarray(y_ret)
        
        # if x is a number (int or float)
        return self.__cdf(x)

    def plot_pdf(self, interval=None, bins=1000, savefig_path=None):
        """Plots the PDF of the Manova Spectrum distribution.

        Args:
            interval (tuple, default=None): Delimiters (xmin, xmax) of the histogram. If not
                provided, the used interval is calculated depending on beta, a, and b.
            bins (int, default=100): It defines the number of equal-width bins within the
                provided interval or range.
            savefig_path (string, default=None): path to save the created figure. If it is not
                provided, the plot is shown at the end of the routine.
        
        """
        if interval is None:
            interval = (self.lambda_minus, self.lambda_plus)
        
        _plot_func(
            interval, func=self.pdf, bins=bins, plot_title="Manova spectrum PDF",
            plot_ylabel="probability density", savefig_path=savefig_path
        )

    def plot_cdf(self, interval=None, bins=1000, savefig_path=None):
        """Plots the CDF of the Manova Spectrum distribution.

        Args:
            interval (tuple, default=None): Delimiters (xmin, xmax) of the histogram. If not
                provided, the used interval is calculated depending on beta, a, and b.
            bins (int, default=100): It defines the number of equal-width bins within the
                provided interval or range.
            savefig_path (string, default=None): path to save the created figure. If it is not
                provided, the plot is shown at the end of the routine.
        
        """
        if interval is None:
            interval = (self.lambda_minus, self.lambda_plus)

        _plot_func(
            interval, func=self.cdf, bins=bins, plot_title="Manova spectrum CDF",
            plot_ylabel="cumulative distribution", savefig_path=savefig_path
        )

    def plot_empirical_pdf(self, m_size=1000, n1_size=None, n2_size=None, bins=100, interval=None,
                           density=False, plot_law_pdf=False, savefig_path=None):
        """Computes and plots Manova spectrum empirical pdf and analytical distribution.

        Calculates and plots Manova spectrum empirical pdf using Manova Ensemble random matrices.
        The size of the sampeld matrices will depend on the `m_size` (:math:`m`) parameter (no. of
        degrees of freedom) and on the ratios :math:`a` and :math:`b` given when instantiating this
        class, unless the parameters `n1_size` (:math:`n_1`) and/or `n2_size` (:math:`n_2`) are also
        given. In this last case, the new ratio :math:`a` for the empirical pdf plotting is computed
        as :math:`a = n_1/m`, and the new ratio :math:`b` is calculated as :math:`b = n_2/m`. If only
        the number of degrees of freedom :math:`m` is provided, the sample sizes :math:`n_1` and :math:`n_2`
        are computed as :math:`n_1 = [a * m]` and :math:`n_2 = [b * m]` respectively.

        Args:
            m_size (int, default=1000): number of rows of the two Wishart Ensemble matrices that
                generates the matrix of the corresponding ensemble. This is the number of degrees
                of freedom (:math:`m`). The sample sizes (:math:`n_1` and :math:`n_2`) of the two
                matrices are computed from this value and the ratios :math:`a` and :math:`b` given
                to instantiate this class.
            n1_size (int, default=None): number of columns of the first Wishart Ensemble matrix
                that generates the matrix of the corresponding ensemble (:math:`n_1`). If provided,
                the ratio :math:`a` is ignored (but not replaced) and the new ratio :math:`a = n_1/m`
                is used instead to plot the empirical pdf.
            n2_size (int, default=None): number of columns of the second Wishart Ensemble matrix
                that generates the matrix of the corresponding ensemble (:math:`n_2`). If provided,
                the ratio :math:`b` is ignored (but not replaced) and the new ratio :math:`b = n_2/m`
                is used instead to plot the empirical pdf.
            bins (int or sequence, default=100): If bins is an integer, it defines the number
                of equal-width bins in the range. If bins is a sequence, it defines the
                bin edges, including the left edge of the first bin and the right
                edge of the last bin; in this case, bins may be unequally spaced.
            interval (tuple, default=None): Delimiters (xmin, xmax) of the histogram.
                The lower and upper range of the bins. Lower and upper outliers are ignored.
            density (bool, default=False): If True, draw and return a probability
                density: each bin will display the bin's raw count divided by the total
                number of counts and the bin width, so that the area under the histogram
                integrates to 1. If set to False, the absolute frequencies of the eigenvalues
                are returned.
            plot_law_pdf (bool, default=False): If True, the limiting theoretical law is plotted.
                If set to False, just the empirical histogram is shown. This parameter is only
                considered when the argument 'density' is set also to True.
            savefig_path (string, default=None): path to save the created figure. If it is not
                provided, the plot is shown are the end of the routine.

        References:
            - Laszlo, L. and Farrel, B.
                "Local Eigenvalue Density for General MANOVA Matrices".
                Journal of Statistical Physics. 152.6 (2013): 1003-1032.
            - Albrecht, J. and Chan, C.P. and Edelman, A.
                "Sturm sequences and random eigenvalue distributions".
                Foundations of Computational Mathematics. 9.4 (2008): 461-483.
            - Dumitriu, I. and Edelman, A.
                "Matrix Models for Beta Ensembles".
                Journal of Mathematical Physics. 43.11 (2002): 5830-5847.

        """
        if m_size<1 or (n1_size is not None and n1_size<1) or (n2_size is not None and n2_size<1):
            raise ValueError("matrix size must be positive")

        if n1_size is None:
            n1_size = round(self.a * m_size)
        
        if n2_size is None:
            n2_size = round(self.b * m_size)

        ens = ManovaEnsemble(beta=self.beta, m=m_size, n1=n1_size, n2=n2_size)

        # computing approximated ratios since the n1_size and n2_size parameters
        # could have been rounded previously
        approx_a = n1_size/m_size
        approx_b = n2_size/m_size
        if approx_a <= 1 or approx_b <= 1:
            print("Warning: sample size ('n1_size' or 'n2_size') too small compared "
                  "to degrees of freedom ('m_size'). It may cause numerical instability.")
        lambda_term1 = np.sqrt((approx_a/(approx_a+approx_b)) * (1 - (1/(approx_a+approx_b))))
        lambda_term2 = np.sqrt((1/(approx_a+approx_b)) * (1 - (approx_a/(approx_a+approx_b))))
        lambda_minus = (lambda_term1 - lambda_term2)**2
        lambda_plus = (lambda_term1 + lambda_term2)**2

        if interval is None:
            interval = [lambda_minus, lambda_plus]
            if approx_a <= 1:
                interval[0] = min(-0.05, lambda_minus)
            if approx_b <= 1:
                interval[1] = max(lambda_plus, 1.05)
            interval = tuple(interval)

        observed, bins = ens.eigval_hist(
            bins=bins,
            interval=interval,
            density=density,
            normalize=True,
            avoid_img=True
        )

        width = bins[1]-bins[0]
        plt.bar(bins[:-1], observed, width=width, align='edge')

        # Plotting theoretical graphic
        if plot_law_pdf and density:
            centers = np.array(_get_bins_centers_and_contour(bins))
            # creating new instance with the approximated ratios depending on the given matrix sizes
            msd = ManovaSpectrumDistribution(beta=self.beta, a=approx_a, b=approx_b)
            pdf = msd.pdf(centers)
            plt.plot(centers, pdf, color='red', linewidth=2)

        plt.title("Manova Spectrum - Eigenvalue histogram", fontweight="bold")
        plt.xlabel("x")
        plt.ylabel("probability density")
        if approx_a <= 1 or approx_b <= 1:
            if plot_law_pdf and density:
                ylim_vals = pdf
            else:
                ylim_vals = observed
            try:
                plt.ylim(0, np.max(ylim_vals) + 0.25*np.max(ylim_vals))
            except ValueError:
                second_highest_val = np.partition(ylim_vals.flatten(), -2)[-2]
                plt.ylim(0, second_highest_val + 0.25*second_highest_val)

        # Saving plot or showing it
        if savefig_path:
            plt.savefig(savefig_path, dpi=1200)
        else:
            plt.show()



"""
Wishart Ensemble
=================

Defines Wishart Ensemble random matrices
"""

# Author: Alejandro Santorum Varela
# License: BSD 3-Clause

from skrmt.ensemble import WishartEnsemble

##############################################################################
# Wishart Ensemble contains random matrices formed by the multiplication
# of two random matrices whose entries are gaussian distributed.

##############################################################################
# Wishart Real Ensemble (WRE)
# ---------------------------
#
# Random matrices of WRE are formed by multiplying a random real standard
# gaussian matrix of size :math:`p \times n` by its transpose.
#
# They are also known as 1-Laguerre random matrices (beta = 1).

##############################################################################
# A random matrix of Wishart Real Ensemble can be sampled using
# scikit-rmt with the following code.

wre = WishartEnsemble(beta=1, p=3, n=5)
print(wre.matrix)

##############################################################################
# Wishart Complex Ensemble (WCE)
# ------------------------------
#
# Random matrices of WCE are formed by multiplying a random complex
# standard gaussian matrix of size :math:`p \times n` by its transpose.
#
# They are also known as 2-Laguerre random matrices (beta = 2).

##############################################################################
# A random matrix of Wishart Complex Ensemble can be sampled using
# scikit-rmt with the following code.

wce = WishartEnsemble(beta=2, p=3, n=5)
print(wce.matrix)

##############################################################################
# Wishart Quaternion Ensemble (WQE)
# ---------------------------------
#
# Random matrices of WQE are formed by sampling two random complex standard
# guassian matrices (:math:`\mathbf{X}` and :math:`\mathbf{Y}`), stacking
# them to create matrix :math:`\mathbf{A} = (\mathbf{X}\ \mathbf{Y}; -\mathbf{Y}^*\ \mathbf{X}^*)`.
# Finally matrix :math:`\mathbf{A}` is multiplied by its transpose to generate
# a matrix WQE randon matrix.
#
# They are also known as 4-Laguerre random matrices (beta = 4).

##############################################################################
# A random matrix of Wishart Quaternion Ensemble can be sampled using
# scikit-rmt with the following code.

wqe = WishartEnsemble(beta=4, p=2, n=5)
print(wqe.matrix)


'''Base Ensemble Test module

Testing _Ensemble abstract class
'''

import os
import pytest
import shutil

from skrmt.ensemble._base_ensemble import _Ensemble
from skrmt.ensemble.gaussian_ensemble import GaussianEnsemble


TMP_DIR_PATH = os.path.join(os.getcwd(), "skrmt/ensemble/tests/tmp")


@pytest.fixture(scope="module", autouse=True)
def _setup_tmp_dir(request):
    '''Function that is run before all tests in this script.

    It creates a temporary folder in order to store useful files
    for the following tests.
    '''
    # if the directory already exists, it is deleted
    if os.path.exists(TMP_DIR_PATH):
        shutil.rmtree(TMP_DIR_PATH)
    # creating temporary directory  
    os.mkdir(TMP_DIR_PATH)

    # specifying a function that will be run after all tests are completed
    request.addfinalizer(_remove_tmp_dir)


def _remove_tmp_dir():
    '''Function that removes the created temporary directory.

    The function is run when all tests in this module are completed.
    '''
    shutil.rmtree(TMP_DIR_PATH)


def test_init_exception():
    """Testing the abstract class cannot be instantiated
    """
    with pytest.raises(TypeError):
        _ = _Ensemble()


def test_base_set_eigval_norm_const():
    """Testing setting a custom eigenvalue normalization constant
    """
    goe = GaussianEnsemble(beta=1, n=10, use_tridiagonal=False)

    assert goe.eigval_norm_const is not None

    goe.set_eigval_norm_const(100.0)
    assert goe.eigval_norm_const == 100.0

    goe.set_eigval_norm_const(0.1)
    assert goe.eigval_norm_const == 0.1



def test_base_ens_plot():
        '''Testing WignerSemicircleDistribution plot pdf
        '''
        fig_name = "test_base_ens_plot_eigval_hist.png"
        goe = GaussianEnsemble(beta=1, n=100, use_tridiagonal=False)
        goe.plot_eigval_hist(fig_path=TMP_DIR_PATH+"/"+fig_name)
        assert os.path.isfile(os.path.join(TMP_DIR_PATH, fig_name)) == True



"""
Introduction
============

In this section, we will briefly explain what is Random Matrix Theory (RMT),
and we will introduce scikit-rmt, a library that provides RMT tools to
mathematicians, engineers, physicists, economists or any interested scientist
in the Python scientific ecosystem.

"""

# Author: Alejandro Santorum Varela
# License: BSD 3-Clause

##############################################################################
# What is Random Matrix Theory?
# ------------------------------
# 
# Random mathematical objects, such as graphs or tensors, have always been
# deeply studied because of their applications in economics, physics and
# engineering. The growing computational capacity allows to work and
# simulate this class of objects efficiently to improve in many disciplines.
# 
# In particular, two-dimensional random tensors are known as random matrices.
# Random Matrix Theory is the is the branch of statistics that studies the
# properties of matrices whose inputs are random variables. Many random matrices
# have similar behaviour and properties, especially if we analyze their
# eigenvalue spectrum, that is why they are usually classified into ensembles
# of random matrices, which are sets of them that share common features.

##############################################################################
# What is scikit-rmt?
# -------------------
#
# scikit-rmt is a Python library containing classes and functions that allow
# you to perform RMT tasks. Using it you can:
#
#     - Sample many types of random matrices of the main ensembles: Gaussian
#       ensemble, Wishart ensemble, Manova ensemble and Circular ensemble.
#     - Plot and analyze random matrix eigenvalue spectrum of the sampled
#       ensembles.
#     - Calculate eigenvalue joint probability density function of the sampled
#       random matrices.
#     - Plot and study main spectrum laws: Wigner Semicircle Law,
#       Marchenko-Pastur Law and Tracy-Widom Law.
#     - Estimation of covariance matrices with several methods, such as
#       non-lineal shrinkage analytical estimator.

##############################################################################
# As an example, the following code shows how to sample the most known
# ensemble of random matrices: Gaussian Orthogonal Ensemble (GOE).

from skrmt.ensemble import GaussianEnsemble

goe = GaussianEnsemble(beta=1, n=5)
print(goe.matrix)

##############################################################################
# To study its eigenvalue spectrum, its size should be a little bit larger.

from skrmt.ensemble import GaussianEnsemble

goe = GaussianEnsemble(beta=1, n=1000)
goe.plot_eigval_hist(bins=80)



"""Manova Ensemble Module

This module contains the implementation of the Manova Ensemble, also
known as Jacobi Ensemble. This ensemble of random matrices contains
mainly three sub-ensembles: Manova Real Ensemble, Manova Complex Ensemble
and Manova Quaternion Ensemble.

"""

import numpy as np
from scipy import special

from ._base_ensemble import _Ensemble


#########################################################################
### Manova Ensemble = Jacobi Ensemble

class ManovaEnsemble(_Ensemble):
    """General Manova Ensemble class.

    This class contains common attributes and methods for all the
    Manova ensembles. It also defines the basic interface to be
    supported by inherited classes. Manova Ensembles are divided in:
    - Manova Real Ensemble (MRE, beta=1): the random matrices of
    this ensemble are formed by sampling two random real standard
    guassian matrices (X and Y) of size m times n1 and m times n2
    respectively. Then, matrix A = (X * X') / (X * X' + Y * Y')
    generates a matrix of the Manova Real Ensemble.
    - Manova Complex Ensemble (MCE, beta=2): the random matrices
    of this ensemble are formed by sampling two random complex
    standard guassian matrices (X and Y) of size m times n1 and
    m times n2 respectively. Then, matrix A = (X * X') / (X * X' + Y * Y')
    generates a matrix of the Manova Complex Ensemble.
    - Manova Quaternion Ensemble (MQE, beta=4): the random matrices
    of this ensemble are formed by: sampling two random complex
    standard guassian matrices (X1 and X2), both of size m times n1.
    Another two random complex standard guassian matrices (Y1 and Y2),
    both of size m times n2, are sampled. They are stacked forming matrices
    X and Y:
    X = [X1  X2; -conj(X2)  conj(X1)]
    Y = [Y1  Y2; -conj(Y2)  conj(Y1)]
    Finally, matrix A = (X * X') / (X * X' + Y * Y') generates a matrix of
    the Manova Quaternion Ensemble of size m times m.

    Attributes:
        matrix (numpy array): instance of the ManovaReal, ManovaComplex
            or ManovaQuaternion random ensembles. If it is an instance
            of ManovaReal or ManovaComplex, the random matrix is of
            size m times m. If it is a ManovaQuaternion, the random matrix
            is of size 2m times 2m.
        beta (int): descriptive integer of the Manova ensemble type.
            For Real beta=1, for Complex beta=2, for Quaternion beta=4.
        m (int): number of rows of the random guassian matrices that
            generates the matrix of the corresponding ensemble.
        n1 (int): number of columns of the first random guassian matrix
            that generates the matrix of the corresponding ensemble.
        n2 (int): number of columns of the second random guassian matrix
            that generates the matrix of the corresponding ensemble.

    References:
        - Erdos, L. and Farrell, B.
            "Local Eigenvalue Density for General MANOVA Matrices".
            Journal of Statistical Physics. 152.6 (2013): 1003-1032.
        - Dumitriu, I. and Edelman, A. "Matrix Models for Beta Ensembles".
            Journal of Mathematical Physics. 43.11 (2002): 5830-5847.

    """

    def __init__(self, beta, m, n1, n2):
        """Constructor for ManovaEnsemble class.

        Initializes an instance of this class with the given parameters.

        Args:
            beta (int): descriptive integer of the Manova ensemble type.
                For Real beta=1, for Complex beta=2, for Quaternion beta=4.
            m (int): number of rows of the random guassian matrices that
                generates the matrix of the corresponding ensemble.
            n1 (int): number of columns of the first random guassian matrix
                that generates the matrix of the corresponding ensemble.
            n2 (int): number of columns of the second random guassian matrix
                that generates the matrix of the corresponding ensemble.

        """
        super().__init__()
        # pylint: disable=invalid-name
        self.m = m
        self.n1 = n1
        self.n2 = n2
        self.beta = beta
        self._eigvals = None
        self.matrix = self.sample()

    def set_size(self, m, n1, n2, resample_mtx=True):
        # pylint: disable=arguments-differ
        """Setter of matrix size.

        Sets the matrix size. Useful if it has been initialized with a different value.

        Args:
            m (int): number of rows of the random guassian matrices that
                generates the matrix of the corresponding ensemble.
            n1 (int): number of columns of the first random guassian matrix
                that generates the matrix of the corresponding ensemble.
            n2 (int): number of columns of the second random guassian matrix
                that generates the matrix of the corresponding ensemble.
            resample_mtx (bool, default=True): If set to True, the ensemble matrix is
                resampled with the new dimensions.

        """
        self.m = m
        self.n1 = n1
        self.n2 = n2
        if resample_mtx:
            self.matrix = self.sample()

    # pylint: disable=inconsistent-return-statements
    def sample(self):
        """Samples new Manova Ensemble random matrix.

        The sampling algorithm depends on the specification of
        beta parameter. If beta=1, Manova Real is sampled; if
        beta=2 Manova Complex is sampled and if beta=4
        Manova Quaternion is sampled.

        Returns:
            numpy array containing new matrix sampled.

        References:
            - Dumitriu, I. and Edelman, A. "Matrix Models for Beta Ensembles".
                Journal of Mathematical Physics. 43.11 (2002): 5830-5847.

        """
        if self.beta == 1:
            return self._sample_mre()
        if self.beta == 2:
            return self._sample_mce()
        if self.beta == 4:
            return self._sample_mqe()

    def _sample_mre(self):
        m_size = self.m
        n1_size = self.n1
        n2_size = self.n2
        # m by n1 random real matrix of random Gaussians
        x_mtx = np.random.randn(m_size,n1_size)
        # m by n2 random real matrix of random Gaussians
        y_mtx = np.random.randn(m_size,n2_size)
        # A1 = X * X'
        a1_mtx = np.matmul(x_mtx, x_mtx.transpose())
        # A2 = X * X' + Y * Y'
        a2_mtx = a1_mtx + np.matmul(y_mtx, y_mtx.transpose())
        # A = (X * X') / (X * X' + Y * Y') = (X * X') * (X * X' + Y * Y')^(-1)
        self.matrix = np.matmul(a1_mtx, np.linalg.inv(a2_mtx))
        # setting array of eigenvalues to None to force re-computing them
        self._eigvals = None
        return self.matrix

    def _sample_mce(self):
        m_size = self.m
        n1_size = self.n1
        n2_size = self.n2
        # m by n1 random complex matrix of random Gaussians
        x_mtx = np.random.randn(m_size,n1_size) + (0+1j)*np.random.randn(m_size,n1_size)
        # m by n2 random complex matrix of random Gaussians
        y_mtx = np.random.randn(m_size,n2_size) + (0+1j)*np.random.randn(m_size,n2_size)
        # A1 = X * X'
        a1_mtx = np.matmul(x_mtx, x_mtx.transpose().conj())
        # A2 = X * X' + Y * Y'
        a2_mtx = a1_mtx + np.matmul(y_mtx, y_mtx.transpose().conj())
        # A = (X * X') / (X * X' + Y * Y') = (X * X') * (X * X' + Y * Y')^(-1)
        self.matrix = np.matmul(a1_mtx, np.linalg.inv(a2_mtx))
        # setting array of eigenvalues to None to force re-computing them
        self._eigvals = None
        return self.matrix

    def _sample_mqe(self):
        m_size = self.m
        n1_size = self.n1
        n2_size = self.n2
        # m by n1 random complex matrix of random Gaussians
        x1_mtx = np.random.randn(m_size,n1_size) + (0+1j)*np.random.randn(m_size,n1_size)
        # m by n1 random complex matrix of random Gaussians
        x2_mtx = np.random.randn(m_size,n1_size) + (0+1j)*np.random.randn(m_size,n1_size)
        # m by n2 random complex matrix of random Gaussians
        y1_mtx = np.random.randn(m_size,n2_size) + (0+1j)*np.random.randn(m_size,n2_size)
        # m by n2 random complex matrix of random Gaussians
        y2_mtx = np.random.randn(m_size,n2_size) + (0+1j)*np.random.randn(m_size,n2_size)
        # X = [X1 X2; -conj(X2) conj(X1)]
        x_mtx = np.block([
                        [x1_mtx               , x2_mtx],
                        [-np.conjugate(x2_mtx), np.conjugate(x1_mtx)]
                        ])
        # Y = [Y1 Y2; -conj(Y2) conj(Y1)]
        y_mtx = np.block([
                         [y1_mtx               , y2_mtx],
                         [-np.conjugate(y2_mtx), np.conjugate(y1_mtx)]
                        ])
        # A1 = X * X'
        a1_mtx = np.matmul(x_mtx, x_mtx.transpose().conj())
        # A2 = X * X' + Y * Y'
        a2_mtx = a1_mtx + np.matmul(y_mtx, y_mtx.transpose().conj())
        # A = (X * X') / (X * X' + Y * Y') = (X * X') * (X * X' + Y * Y')^(-1)
        self.matrix = np.matmul(a1_mtx, np.linalg.inv(a2_mtx))
        # setting array of eigenvalues to None to force re-computing them
        self._eigvals = None
        return self.matrix

    def eigvals(self, normalize=False):
        """Computes the random matrix eigenvalues.

        Calculates the random matrix eigenvalues using numpy standard procedure.
        If the matrix ensemble is symmetric, a faster algorithm is used.

        Returns:
            numpy array with the calculated eigenvalues.

        """
        norm_const = self.eigval_norm_const if normalize else 1.0

        if self._eigvals is not None:
            return norm_const * self._eigvals

        # always storing non-normalized eigenvalues
        self._eigvals = np.linalg.eigvals(self.matrix)
        return norm_const * self._eigvals

    def plot_eigval_hist(self, bins, interval=(0,1), density=False, normalize=True, fig_path=None):
        """Computes and plots the histogram of the matrix eigenvalues

        Calculates and plots the histogram of the current sampled matrix eigenvalues.

        Args:
            bins (int or sequence): If bins is an integer, it defines the number of
                equal-width bins in the range. If bins is a sequence, it defines the
                bin edges, including the left edge of the first bin and the right
                edge of the last bin; in this case, bins may be unequally spaced.
            interval (tuple, default=(0,1)): Delimiters (xmin, xmax) of the histogram.
                The lower and upper range of the bins. Lower and upper outliers are ignored.
            density (bool, default=False): If True, draw and return a probability
                density: each bin will display the bin's raw count divided by the total
                number of counts and the bin width, so that the area under the histogram
                integrates to 1. If set to False, the absolute frequencies of the eigenvalues
                are returned.
            normalize (bool, default=True): Whether to normalize the computed eigenvalues
                by the default normalization constant (see references). Defaults to True, i.e.,
                the eigenvalues are normalized. Normalization makes the eigenvalues to be in the
                same support independently of the sample size.
            fig_path (string, default=None): path to save the created figure. If it is not
                provided, the plot is shown at the end of the routine.

        References:
            - Erdos, L. and Farrell, B.
                "Local Eigenvalue Density for General MANOVA Matrices".
                Journal of Statistical Physics. 152.6 (2013): 1003-1032.
            - Dumitriu, I. and Edelman, A.
                "Matrix Models for Beta Ensembles".
                Journal of Mathematical Physics. 43.11 (2002): 5830-5847.

        """
        # pylint: disable=too-many-arguments
        return super().plot_eigval_hist(
            bins=bins,
            interval=interval,
            density=density,
            normalize=normalize,
            fig_path=fig_path,
            avoid_img=True,
        )

    def joint_eigval_pdf(self, eigvals=None):
        '''Computes joint eigenvalue pdf.

        Calculates joint eigenvalue probability density function given an array of
        eigenvalues. If the array of eigenvalues is not provided, the current random
        matrix sample (so its eigenvalues) is used. This function depends on beta,
        i.e., in the sub-Manova ensemble.

        Args:
            eigvals (np.ndarray, default=None): numpy array with the values (eigenvalues)
                to evaluate the joint pdf in.

        Returns:
            real number. Value of the joint pdf of the eigenvalues.

        References:
            - Dumitriu, I. and Edelman, A. "Matrix Models for Beta Ensembles".
                Journal of Mathematical Physics. 43.11 (2002): 5830-5847.

        '''
        # pylint: disable=invalid-name
        a1 = self.beta*self.n1/2
        a2 = self.beta*self.n2/2
        p = 1 + self.beta/2*(self.m - 1)

        if eigvals is None:
            # calculating eigenvalues
            eigvals = self.eigvals()
        n_eigvals = len(eigvals)

        # calculating Jacobi eigval pdf constant depeding on beta
        const_beta = 1
        for j in range(self.m):
            const_beta *= (special.gamma(1 + self.beta/2) * \
                            special.gamma(a1 + a2 -self.beta/2*(self.m - j)))/ \
                          (special.gamma(1 + self.beta*j/2) * \
                            special.gamma(a1 - self.beta/2*(self.m - j)) * \
                              special.gamma(a2 - self.beta/2*(self.m - j)))

        # calculating first prod
        prod1 = 1
        for j in range(n_eigvals):
            for i in range(j):
                prod1 *= np.abs(eigvals[i] - eigvals[j])**self.beta
        # calculating second prod
        prod2 = 1
        for j in range(n_eigvals):
            prod2 *= eigvals[j]**(a1-p) * (1 - eigvals[j])**(a2-p)
        # calculating Jacobi eigval pdf
        return const_beta * prod1 * prod2



"""
Manova Ensemble
=================

Defines Manova Ensemble random matrices
"""

# Author: Alejandro Santorum Varela
# License: BSD 3-Clause

from skrmt.ensemble import ManovaEnsemble

##############################################################################
# Manova Ensemble random matrices are considered to be double Wishart random
# matrices.

##############################################################################
# Manova Real Ensemble (MRE)
# ---------------------------
#
# Random matrices of MRE are formed by sampling two random real standard
# guassian matrices (:math:`\mathbf{X}` and :math:`\mathbf{Y}`) of size
# :math:`m \times n_1` and :math:`m \times n_2` respectively. Then, matrix
# :math:`\mathbf{A} = \dfrac{\mathbf{X}\mathbf{X}^T}{\mathbf{X}\mathbf{X}^T  + \mathbf{Y}\mathbf{Y}^T}`
# generates a matrix of the MRE.
#
# They are also known as 1-Jacobi random matrices (beta = 1).

##############################################################################
# A random matrix of Manova Real Ensemble can be sampled using
# scikit-rmt with the following code.

mre = ManovaEnsemble(beta=1, m=3, n1=5, n2=5)
print(mre.matrix)

##############################################################################
# Manova Complex Ensemble (MCE)
# ------------------------------
#
# Random matrices of MCE are formed by sampling two random complex standard
# guassian matrices (:math:`\mathbf{X}` and :math:`\mathbf{Y}`) of size
# :math:`m \times n_1` and :math:`m \times n_2` respectively. Then, matrix
# :math:`\mathbf{A} = \dfrac{\mathbf{X}\mathbf{X}^T}{\mathbf{X}\mathbf{X}^T  + \mathbf{Y}\mathbf{Y}^T}`
# generates a matrix of the MCE.
#
# They are also known as 2-Jacobi random matrices (beta = 2).

##############################################################################
# A random matrix of Manova Complex Ensemble can be sampled using
# scikit-rmt with the following code.

mce = ManovaEnsemble(beta=2, m=3, n1=5, n2=5)
print(mce.matrix)

##############################################################################
# Manova Quaternion Ensemble (MQE)
# ---------------------------------
#
# Random matrices of MQE are formed by sampling two random complex standard 
# guassian matrices (:math:`\mathbf{X_1}` and :math:`\mathbf{X_2}`), both of
# size :math:`m \times n_1`. Another two random complex standard guassian matrices
# (:math:`\mathbf{Y_1}` and :math:`\mathbf{Y_2}`), both of size :math:`m \times n_2`,
# are sampled. They are stacked forming matrices :math:`\mathbf{X}` and
# :math:`\mathbf{Y}`:
# 
# :math:`\mathbf{X} = (\mathbf{X_1}\ \mathbf{X_2}; -\mathbf{X_2}^*\ \mathbf{X_1}^*)`
# 
# :math:`\mathbf{Y} = (\mathbf{Y_1}\ \mathbf{Y_2}; -\mathbf{Y_2}^*\ \mathbf{Y_1}^*)`
#
# Finally, matrix :math:`\mathbf{A} = \dfrac{\mathbf{X}\mathbf{X}^T}{\mathbf{X}\mathbf{X}^T  + \mathbf{Y}\mathbf{Y}^T}`
# generates a matrix of the MQE.
#
# They are also known as 4-Jacobi random matrices (beta = 4).

##############################################################################
# A random matrix of Manova Quaternion Ensemble can be sampled using
# scikit-rmt with the following code.

mqe = ManovaEnsemble(beta=4, m=2, n1=5, n2=5)
print(mqe.matrix)


"""Base Ensemble Module

This module contains the general implementation of the matrix ensembles.
This file contains the common attributes and methods for all the
random matrix ensembles. It also defines the basic interface to be
supported by inherited classes.

"""

from abc import ABCMeta, abstractmethod

import numpy as np
import matplotlib.pyplot as plt


#########################################################################
### ABSTRACT CLASS: ENSEMBLE

class _Ensemble(metaclass=ABCMeta):
    """General abstract ensemble class.

    This class contains common attributes and methods for all the
    ensembles. It also defines the basic interface to be
    supported by inherited classes.

    Attributes:
        matrix (numpy array): instance of the random matrix ensemble
            of size n times n.
        _eigvals (numpy array): array of computed eigenvalues. This array
            is None until the method `eigvals` is called. The computed
            eigenvalues are cached in the attribute _eigvals to avoid
            re-computing them. The eigenvalues are re-calculated again
            if the matrix sample changes.
        eigval_norm_cost (float): constant that is used to normalize the
            eigenvalues so their support is always the same independently
            of the sample size. This is really useful for plotting, since
            the representation interval will always be the same no matter
            how many eigenvalues are sampled.
    """

    @abstractmethod
    def __init__(self):
        self.matrix = None
        self._eigvals = None
        # default eigenvalue normalization constant
        self.eigval_norm_const = 1.0

    @abstractmethod
    def sample(self):
        """Samples new random matrix.

        The sampling algorithm depends on the inherited classes, so it should be
        specified by them.

        Returns:
            numpy array containing new matrix sampled.
        """
        # pylint: disable=unnecessary-pass
        pass

    @abstractmethod
    def set_size(self):
        # pylint: disable=unnecessary-pass
        # pylint: disable=missing-function-docstring
        # this will be commented at inherited classes
        pass

    def set_eigval_norm_const(self, eigval_norm_const):
        """Sets a custom eigenvalue normalization constant.

        This updates the normalization constant applied to the computed eigenvalues.
        Eigenvalue normalization is useful because normalized eigenvalues always have
        the same support independently of the sample size.

        Args:
            eigval_norm_const (float): new eigenvalue normalization constant.
        """
        # pylint: disable=unnecessary-pass
        self.eigval_norm_const = eigval_norm_const

    @abstractmethod
    def eigvals(self, normalize=False):
        # pylint: disable=unnecessary-pass
        # pylint: disable=missing-function-docstring
        # this will be commented at inherited classes
        pass

    @abstractmethod
    def joint_eigval_pdf(self):
        # pylint: disable=unnecessary-pass
        # pylint: disable=missing-function-docstring
        # this will be commented at inherited classes
        pass

    def eigval_hist(
        self,
        bins,
        interval=None,
        density=False,
        normalize=True,
        avoid_img=False
    ):
        """Calculates the histogram of the matrix eigenvalues.

        Calculates the histogram of the current sampled matrix eigenvalues. Some ensembles
        like Gaussian (Hermite) ensemble or Wishart (Laguerre) ensemble might have
        improved routines to avoid calculating the eigenvalues, so instead the histogram
        is built using certain techniques to boost efficiency.
        It is important to underline that this function works with real eigenvalues,
        if the matrix eigenvalues are complex, they are casted to its real part.

        Args:
            bins (int or sequence): If bins is an integer, it defines the number of
                equal-width bins in the range. If bins is a sequence, it defines the
                bin edges, including the left edge of the first bin and the right
                edge of the last bin; in this case, bins may be unequally spaced.
            interval (tuple): Delimiters (xmin, xmax) of the histogram.
                The lower and upper range of the bins. Lower and upper outliers are ignored.
            density (bool, default=False): If True, draw and return a probability
                density: each bin will display the bin's raw count divided by the total
                number of counts and the bin width, so that the area under the histogram
                integrates to 1. If set to False, the absolute frequencies of the eigenvalues
                are returned.
            normalize (bool, default=True): Whether to normalize the computed eigenvalues
                by the default normalization constant (see references). Defaults to True, i.e.,
                the eigenvalues are normalized. Normalization makes the eigenvalues to be in the
                same support independently of the sample size.
            avoid_img (bool, default=False): If True, eigenvalue imaginary part is ignored.
                This should be used when the eigenvalue compatation is expected to generate
                complex eigenvalues with really small imaginary part because of computing
                rounding errors. E.g.: MANOVA Ensemble eigenvalues.

        Returns:
            (tuple) tuple containing:
                observed (array): List of eigenvalues frequencies per bin. If density is
                True these values are the relative frequencies in order to get an area under
                the histogram equal to 1. Otherwise, this list contains the absolute
                frequencies of the eigenvalues.
                bins (array): The edges of the bins. Length nbins + 1 (nbins left edges and
                right edge of last bin)

        Raises:
            ValueError if interval is not a tuple.

        References:
            - Albrecht, J. and Chan, C.P. and Edelman, A.
                "Sturm sequences and random eigenvalue distributions".
                Foundations of Computational Mathematics. 9.4 (2008): 461-483.
            - Dumitriu, I. and Edelman, A.
                "Matrix Models for Beta Ensembles".
                Journal of Mathematical Physics. 43.11 (2002): 5830-5847.

        """
        if interval is not None:
            if not isinstance(interval, tuple):
                raise ValueError("interval argument must be a tuple")

        # calculating eigenvalues using standard algorithm
        eigvals = self.eigvals(normalize=normalize)
        # ignoring imaginary part because of computing rounding errors
        if avoid_img:
            eigvals = eigvals.real

        # using numpy to obtain histogram in the given interval and no. of bins
        observed, bins = np.histogram(eigvals, bins=bins, range=interval, density=density)
        return observed, bins


    def plot_eigval_hist(
        self,
        bins,
        interval=None,
        density=False,
        normalize=True,
        fig_path=None,
        avoid_img=False
    ):
        """Computes and plots the histogram of the matrix eigenvalues.

        Calculates and plots the histogram of the current sampled matrix eigenvalues.
        Some ensembles like Gaussian (Hermite) ensemble or Wishart (Laguerre) ensemble
        have improved routines to avoid calculating the eigenvalues, so the histogram
        is built using certain techniques to boost efficiency. It is important to
        underline that this function works with real and complex eigenvalues: if the
        matrix eigenvalues are complex, they are plotted in the complex plane next to a
        heap map to study eigenvalue density.

        Args:
            bins (int or sequence): If bins is an integer, it defines the number of
                equal-width bins in the range. If bins is a sequence, it defines the
                bin edges, including the left edge of the first bin and the right
                edge of the last bin; in this case, bins may be unequally spaced.
            interval (tuple): Delimiters (xmin, xmax) of the histogram.
                The lower and upper range of the bins. Lower and upper outliers are ignored.
            density (bool, default=False): If True, draw and return a probability
                density: each bin will display the bin's raw count divided by the total
                number of counts and the bin width, so that the area under the histogram
                integrates to 1. If set to False, the absolute frequencies of the eigenvalues
                are returned.
            normalize (bool, default=True): Whether to normalize the computed eigenvalues
                by the default normalization constant (see references). Defaults to True, i.e.,
                the eigenvalues are normalized. Normalization makes the eigenvalues to be in the
                same support independently of the sample size.
            avoid_img (bool, default=False): If True, eigenvalue imaginary part is ignored.
                This should be used when the eigenvalue compatation is expected to generate
                complex eigenvalues with really small imaginary part because of computing
                rounding errors. E.g.: MANOVA Ensemble eigenvalues.
            fig_path (string, default=None): path to save the created figure. If it is not
                provided, the plot is shown at the end of the routine.

        References:
            - Albrecht, J. and Chan, C.P. and Edelman, A.
                "Sturm sequences and random eigenvalue distributions".
                Foundations of Computational Mathematics. 9.4 (2008): 461-483.
            - Dumitriu, I. and Edelman, A.
                "Matrix Models for Beta Ensembles".
                Journal of Mathematical Physics. 43.11 (2002): 5830-5847.

        """
        # pylint: disable=too-many-arguments
        if not isinstance(interval, tuple):
            raise ValueError("interval argument must be a tuple")

        observed, bins = self.eigval_hist(bins=bins, interval=interval, density=density,
                                          normalize=normalize, avoid_img=avoid_img)
        width = bins[1]-bins[0]
        plt.bar(bins[:-1], observed, width=width, align='edge')

        plt.title("Eigenvalue histogram", fontweight="bold")
        plt.xlabel("x")
        plt.ylabel("density")

        # Saving plot or showing it
        if fig_path:
            plt.savefig(fig_path, dpi=1200)
        else:
            plt.show()



'''Tracy-Widom distribution approximator Test module

Testing module that checks Tracy-Widom distribution approximator.
'''

from numpy.testing import assert_almost_equal

from skrmt.ensemble.tracy_widom_approximator import TW_Approximator

def test_cdf_approximation():
    """Testing the approximation of the CDF.
    
    The PDF is not tested since it is the derivative of the CDF.
    """
    tw = TW_Approximator(beta=2)

    assert_almost_equal(tw.cdf(0.1), 0.9754704606594619, decimal=5)

    assert_almost_equal(tw.cdf(-1.0), 0.8072142419992853, decimal=5)
    assert_almost_equal(tw.cdf(-2.0), 0.41322414250512257, decimal=4)
    assert_almost_equal(tw.cdf(-3.0), 0.08031955293933454, decimal=4)
    assert_almost_equal(tw.cdf(-4.0), 0.0035445535955092003, decimal=4)
    assert_almost_equal(tw.cdf(-5.0), 2.135996984741116e-5, decimal=4)
    assert_almost_equal(tw.cdf(-6.0), 1.062254674124451e-8, decimal=4)
    assert_almost_equal(tw.cdf(-7.0), 2.639614767246062e-13, decimal=4)
    assert_almost_equal(tw.cdf(-8.0), 1.9859004257636574e-19, decimal=4)

    assert_almost_equal(tw.cdf(1.0), 0.9975054381493893, decimal=4)
    assert_almost_equal(tw.cdf(2.0), 0.9998875536983092, decimal=4)
    assert_almost_equal(tw.cdf(3.0), 0.9999970059566077, decimal=5)
    assert_almost_equal(tw.cdf(4.0), 0.9999999504208784, decimal=6)
    assert_almost_equal(tw.cdf(5.0), 0.9999999994682207, decimal=7)
    assert_almost_equal(tw.cdf(6.0), 0.9999999999961827, decimal=8)
    assert_almost_equal(tw.cdf(7.0), 0.9999999999999811, decimal=9)
    assert_almost_equal(tw.cdf(8.0), 0.9999999999999999, decimal=10)



"""Metrics Module

This module contains the implementation of several metrics to
measure estimators quality.
"""

import numpy as np


def loss_mv(sigma_tilde, sigma):
    """Computes minimum variance loss function.

    Computes minimum variance loss function between estimated covariance
    matrix (Sigma tilde) and population covariance matrix (Sigma).

    Args:
        sigma_tilde (numpy array): estimated covariance matrix.
        sigma (numpy array): population covariance matrix.

    Returns:
        loss (double): calculated minimum variance loss.

    References:
        Ledoit, O. and Wolf, M.
            "Analytical nonlinear shrinkage of large-dimensional covariance matrices".
            Annals of Statistics. 48.5 (2020): 3043-3065

    """
    p_size = len(sigma)

    # pylint: disable=raise-missing-from
    try:
        sigma_tilde_inv = np.linalg.inv(sigma_tilde)
    except np.linalg.LinAlgError:
        raise ValueError("Unable to invert estimated covariance matrix")

    try:
        sigma_inv = np.linalg.inv(sigma)
    except np.linalg.LinAlgError:
        raise ValueError("Unable to invert population covariance matrix")


    mtx = np.matmul(np.matmul(sigma_tilde_inv, sigma), sigma_tilde_inv)
    loss = np.trace(mtx)/p_size/(np.trace(sigma_tilde_inv)/p_size)**2 - p_size/np.trace(sigma_inv)
    return loss


def loss_frobenius(sigma_tilde, sigma):
    """Computes Fröbenius loss function.

    Computes Fröbenius loss function between estimated covariance
    matrix (Sigma tilde) and population covariance matrix (Sigma).

    Args:
        sigma_tilde (numpy array): estimated covariance matrix.
        sigma (numpy array): population covariance matrix.

    Returns:
        loss (double): calculated Fröbenius loss.

    References:
        Ledoit, O. and Wolf, M.
            "Analytical nonlinear shrinkage of large-dimensional covariance matrices".
            Annals of Statistics. 48.5 (2020): 3043-3065

    """
    p_size = len(sigma)
    loss = np.trace(sigma_tilde - sigma)**2/p_size
    return loss


def prial_mv(exp_sample, exp_sigma_tilde, exp_fsopt):
    """Computes percentage relative improvement in average loss.

    Computes percentage relative improvement in average loss using
    minimum variance losses. The given expectations must have been
    calculated using Monte Carlo simulations.

    Args:
        exp_sample (double): expected MV loss between sample covariance
            matrix and population matrix.
        exp_sigma_tilde (double): expected MV loss between estimated covariance
            matrix and population matrix.
        exp_fsopt (double): expected MV loss between S^* covariance
            matrix (i.e., finite-sample optimal covariance matrix)
            and population matrix.

    Returns:
        prial (double): calculated percentage relative improvement in average loss.

    References:
        Ledoit, O. and Wolf, M.
            "Analytical nonlinear shrinkage of large-dimensional covariance matrices".
            Annals of Statistics. 48.5 (2020): 3043-3065

    """
    return (exp_sample - exp_sigma_tilde)/(exp_sample - exp_fsopt)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import re
import sys
sys.path.insert(0, os.path.abspath('.'))
sys.path.append('skrmt')


# -- Project information -----------------------------------------------------

project = 'scikit-rmt'
copyright = '2021, Alejandro Santorum Varela'
author = 'Alejandro Santorum Varela'

# The full version, including alpha/beta/rc tags
VERSION_FILEPATH = "skrmt/_version.py"
with open(
    os.path.join(os.path.dirname(__file__), VERSION_FILEPATH),
    "r",
) as version_file:
    version_file_text = version_file.read()
    VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
    match = re.search(VSRE, version_file_text, re.M)
    if match:
        release = match.group(1)
    else:
        raise RuntimeError(f"Unable to find version in {VERSION_FILEPATH}.")

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'sphinx.ext.autosummary',
    'sphinx_gallery.gen_gallery'
]

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = False
napoleon_use_rtype = False
napoleon_type_aliases = None
napoleon_attr_annotations = True

# Add any paths that contain templates here, relative to this directory.
#templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', 'venv/*']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
#html_static_path = ['_static']

# tutorial_list = [
#     "plot_introduction.py",
# ]

sphinx_gallery_conf = {
    'examples_dirs': ['examples', 'tutorial'],
    'gallery_dirs': ['auto_examples', 'auto_tutorial'],
}


'''Metrics Test module

Testing Metrics module
'''

import numpy as np
import pytest

from skrmt.covariance import sample_estimator
from skrmt.covariance import fsopt_estimator
from skrmt.covariance import loss_frobenius, loss_mv, prial_mv



def test_prial_sample():
    '''Testing prial evaluated in sample covariance matrix
    '''
    # population covariance matrix
    sigma = np.array([[3.00407916, -1.46190757, 1.50140806, 1.50933526, 0.27036442],
                      [-1.46190757, 5.61441061, -0.51939653, -2.76492235, 1.38225566],
                      [1.50140806, -0.51939653, 2.3068582, 1.41248896, 0.84740175],
                      [1.50933526, -2.76492235, 1.41248896, 6.57182938, 0.73407095],
                      [0.27036442, 1.38225566, 0.84740175, 0.73407095, 9.50282265]])

    p_size, n_size = sigma.shape[0], 3*sigma.shape[0]
    # input data matrix
    mtx = np.random.multivariate_normal(np.random.randn(p_size), sigma, size=n_size)

    sigma_sample = sample_estimator(mtx)
    sigma_fsopt = fsopt_estimator(mtx, sigma)

    exp_sample = loss_mv(sigma_tilde=sigma_sample, sigma=sigma)
    exp_sigma_tilde = loss_mv(sigma_tilde=sigma_sample, sigma=sigma)
    exp_fsopt = loss_mv(sigma_tilde=sigma_fsopt, sigma=sigma)

    prial = prial_mv(exp_sample=exp_sample, exp_sigma_tilde=exp_sigma_tilde, exp_fsopt=exp_fsopt)

    assert prial == 0.0


def test_prial_fsopt():
    '''Testing prial evaluated in finite-sample optimal covariance matrix
    '''
    # population covariance matrix
    sigma = np.array([[3.00407916, -1.46190757, 1.50140806, 1.50933526, 0.27036442],
                      [-1.46190757, 5.61441061, -0.51939653, -2.76492235, 1.38225566],
                      [1.50140806, -0.51939653, 2.3068582, 1.41248896, 0.84740175],
                      [1.50933526, -2.76492235, 1.41248896, 6.57182938, 0.73407095],
                      [0.27036442, 1.38225566, 0.84740175, 0.73407095, 9.50282265]])

    p_size, n_size = sigma.shape[0], 3*sigma.shape[0]
    # input data matrix
    mtx = np.random.multivariate_normal(np.random.randn(p_size), sigma, size=n_size)

    sigma_sample = sample_estimator(mtx)
    sigma_fsopt = fsopt_estimator(mtx, sigma)

    exp_sample = loss_mv(sigma_tilde=sigma_sample, sigma=sigma)
    exp_sigma_tilde = loss_mv(sigma_tilde=sigma_fsopt, sigma=sigma)
    exp_fsopt = loss_mv(sigma_tilde=sigma_fsopt, sigma=sigma)

    prial = prial_mv(exp_sample=exp_sample, exp_sigma_tilde=exp_sigma_tilde, exp_fsopt=exp_fsopt)

    assert prial == 1.0


def test_loss_frobenius():
    '''Testing Frobenius loss is zero when applied at the same matrix
    '''
    # population covariance matrix
    sigma = np.array([[3.00407916, -1.46190757, 1.50140806, 1.50933526, 0.27036442],
                      [-1.46190757, 5.61441061, -0.51939653, -2.76492235, 1.38225566],
                      [1.50140806, -0.51939653, 2.3068582, 1.41248896, 0.84740175],
                      [1.50933526, -2.76492235, 1.41248896, 6.57182938, 0.73407095],
                      [0.27036442, 1.38225566, 0.84740175, 0.73407095, 9.50282265]])

    assert loss_frobenius(sigma, sigma) == 0


def test_exceptions():
    '''Testing loss function raises exception when a non-invertible matrix is introduced
    '''
    mtx1 = np.zeros((5, 5))
    mtx2 = np.array([[3.00407916, -1.46190757, 1.50140806, 1.50933526, 0.27036442],
                     [-1.46190757, 5.61441061, -0.51939653, -2.76492235, 1.38225566],
                     [1.50140806, -0.51939653, 2.3068582, 1.41248896, 0.84740175],
                     [1.50933526, -2.76492235, 1.41248896, 6.57182938, 0.73407095],
                     [0.27036442, 1.38225566, 0.84740175, 0.73407095, 9.50282265]])

    with pytest.raises(ValueError):
        loss_mv(mtx1, mtx2)

    with pytest.raises(ValueError):
        loss_mv(mtx2, mtx1)



"""
Complex histograms
==================

Shows how spectrum of random matrices with complex eigenvalues is plotted.
"""

# Author: Alejandro Santorum Varela
# License: BSD 3-Clause

from skrmt.ensemble import CircularEnsemble

##############################################################################
# Several ensembles have symmetric or hermitian matrices, so their eigenvalues
# are real. However, there are some cases where random matrices are not
# symmetric either hermitian. Therefore, their eigenvalues can be complex.
#
# This is the case of Circular Unitary Ensemble (CUE) and Circular Symplectic
# Ensemble (CSE).
#
# scikit-rmt takes into account this possibility and uses histograms in 2D in
# order to represent spectrum distribution of given ensembles.

##############################################################################
# In this example, eigenvalue spectrum of CUE random matrix is shown.

cue = CircularEnsemble(beta=2, n=1000)
cue.plot_eigval_hist(bins=80, interval=(-2.2,2.2))

##############################################################################
# And the eigenvalue spectrum of CSE.

cse = CircularEnsemble(beta=4, n=1000)
cse.plot_eigval_hist(bins=80, interval=(-2.2,2.2))

##############################################################################
# The provided heatmap gives an illustration of the eigenvalue accumulation.





"""
Circular Ensemble
=================

Defines Circular Ensemble random matrices
"""

# Author: Alejandro Santorum Varela
# License: BSD 3-Clause

from skrmt.ensemble import CircularEnsemble

##############################################################################
# Circular Ensemble contains random matrices introduced by Freeman Dyson as
# modifications of the Gaussian Ensemble.

##############################################################################
# Circular Orthogonal Ensemble (COE)
# ----------------------------------
#
# Random matrices of COE have real entries gaussian distributed. These
# matrices are invariant under orthogonal conjugation, i.e., if
# :math:`\mathbf{X} \in \text{COE}` and :math:`\mathbf{O}` is an orthogonal
# matrix, then :math:`\mathbf{O} \mathbf{X} \mathbf{O}^T` is equally
# distributed as :math:`\mathbf{X}`.
#
# They are also known as 1-Dyson random matrices (beta = 1).

##############################################################################
# A random matrix of Circular Orthogonal Ensemble can be sampled using
# scikit-rmt with the following code.

coe = CircularEnsemble(beta=1, n=4)
print(coe.matrix)

##############################################################################
# Circular Unitary Ensemble (CUE)
# -------------------------------
#
# Random matrices of CUE have complex entries gaussian distributed, i.e.,
# their real part and their complex part are gaussian distributed. These
# matrices are invariant under unitary conjugation, i.e., if
# :math:`\mathbf{X} \in \text{CUE}` and :math:`\mathbf{O}` is an unitary
# matrix, then :math:`\mathbf{O} \mathbf{X} \mathbf{O}^T` is equally
# distributed as :math:`\mathbf{X}`.
#
# They are also known as 2-Dyson random matrices (beta = 2).

##############################################################################
# A random matrix of Circular Unitary Ensemble can be sampled using
# scikit-rmt with the following code.

cue = CircularEnsemble(beta=2, n=4)
print(cue.matrix)

##############################################################################
# Circular Symplectic Ensemble (CSE)
# ----------------------------------
#
# Random matrices of CSE  are invariant under conjugation by the symplectic
# group.
#
# They are also known as 4-Dyson random matrices (beta = 4).

##############################################################################
# A random matrix of Circular Symplectic Ensemble can be sampled using
# scikit-rmt with the following code.

cse = CircularEnsemble(beta=4, n=2)
print(cse.matrix)


"""
Gaussian Ensemble
=================

Defines Gaussian Ensemble random matrices
"""

# Author: Alejandro Santorum Varela
# License: BSD 3-Clause

from skrmt.ensemble import GaussianEnsemble

##############################################################################
# Gaussian Ensemble contains random matrices whose inputs are gaussian
# distributed.

##############################################################################
# Gaussian Orthogonal Ensemble (GOE)
# ----------------------------------
#
# Random matrices of GOE have real entries gaussian distributed. These
# matrices are invariant under orthogonal conjugation, i.e., if
# :math:`\mathbf{X} \in \text{GOE}` and :math:`\mathbf{O}` is an orthogonal
# matrix, then :math:`\mathbf{O} \mathbf{X} \mathbf{O}^T` is equally
# distributed as :math:`\mathbf{X}`.
#
# They are also known as 1-Hermite random matrices (beta = 1).

##############################################################################
# A random matrix of Gaussian Orthogonal Ensemble can be sampled using
# scikit-rmt with the following code.

goe = GaussianEnsemble(beta=1, n=4)
print(goe.matrix)

##############################################################################
# Gaussian Unitary Ensemble (GUE)
# -------------------------------
#
# Random matrices of GUE have complex entries gaussian distributed, i.e.,
# their real part and their complex part are gaussian distributed. These
# matrices are invariant under unitary conjugation, i.e., if
# :math:`\mathbf{X} \in \text{GUE}` and :math:`\mathbf{O}` is an unitary
# matrix, then :math:`\mathbf{O} \mathbf{X} \mathbf{O}^T` is equally
# distributed as :math:`\mathbf{X}`.
#
# They are also known as 2-Hermite random matrices (beta = 2).

##############################################################################
# A random matrix of Gaussian Unitary Ensemble can be sampled using
# scikit-rmt with the following code.

gue = GaussianEnsemble(beta=2, n=4)
print(gue.matrix)

##############################################################################
# Gaussian Symplectic Ensemble (GSE)
# ----------------------------------
#
# Random matrices of GSE have quaternionic entries gaussian distributed, i.e.,
# their four dimensions are gaussian distributed. These matrices are invariant
# under conjugation by the symplectic group.
#
# They are also known as 4-Hermite random matrices (beta = 4).

##############################################################################
# A random matrix of Gaussian Symplectic Ensemble can be sampled using
# scikit-rmt with the following code.

gse = GaussianEnsemble(beta=4, n=2)
print(gse.matrix)


"""
Random matrix spectrum distribution
===================================

In this section we describe the tools provided by **scikit-rmt** to study,
analyze and compute the distribution of the spectral laws of some common
random matrix ensembles.

"""

# Author: Alejandro Santorum Varela
# License: BSD 3-Clause


##############################################################################
# Spectral laws
# -------------
# 
# Spectral laws define random matrix eigenvalue distribution when its size
# goes to infinity. So the spectral laws describes the limiting behavior
# of the spectrum of a random matrix ensemble.
#
# The most known spectral laws, and implemented by scikit-rmt, are the
# following:
#
# - **Winger Semicircle Law**: describes the limiting behaviour of Gaussian
#   Ensemble random matrix spectrum.
#
# - **Marchenko-Pastur Law**: describes the limiting behaviour of Wishart
#   Ensemble random matrix spectrum.
#
# - **Tracy-Widom Law**: describes the limiting behaviour of the largest
#   eigenvalue of a Gaussian Ensemble random matrix.
#
# - **Manova Spectrum distribution**: Although the random matrices of the
#   Manova ensemble have a defined limiting behaviour, the analytical law
#   that describes it is not named after anyone like the previous mentioned
#   laws. It was introduced by Wachter in *The Limiting Empirical Measure
#   of Multiple Discriminant Ratios* (1980).


##############################################################################
# Wigner Semicircle Law
# ---------------------
#
# **Wigner's Semicircle Law** characterizes the density of eigenvalues of
# sufficiently large Wigner matrices with second moment :math:`\rho`.
#
#   .. math:: \mu_{sc}(dx) = \frac{1}{2\pi \rho} \sqrt{4\rho - x^2}\mathbf{1}_{|x| \le 2\sqrt{\rho}} dx.
#
# The previous equation can be re-formulated by using that the radius of the
# semicircle is :math:`R = 2\sqrt{\rho}`:
#
#   .. math:: \mu_{sc}(dx) = \frac{2}{\pi R^2} \sqrt{R^2 - x^2}\mathbf{1}_{|x| \le R} dx.
#
# Random matrices from a Gaussian ensemble are of the Wigner type. Therefore,
# distribution of their eigenvalues scaled by :math:`1 / \sqrt{n}`, where
# :math:`n` is the size of the matrix, approaches asymptotically Wigner's Law
# at rate :math:`O(n^{-1/2})`.
#
# The *probability density function* (PDF) of the Wigner's Semicircle Law can
# be studied and plotted using **scikit-rmt** with the
# **WignerSemicircleDistribution** class.

import numpy as np
import matplotlib.pyplot as plt
from skrmt.ensemble.spectral_law import WignerSemicircleDistribution


x1 = np.linspace(-5, 5, num=1000)
x2 = np.linspace(-10, 10, num=2000)

fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(12,4))

for sigma in [0.5, 1.0, 2.0, 4.0]:
    wsd = WignerSemicircleDistribution(beta=1, center=0.0, sigma=sigma)

    y1 = wsd.pdf(x1)
    y2 = wsd.pdf(x2)

    ax1.plot(x1, y1, label=f"$\sigma$ = {sigma} (R = ${wsd.radius}$)")
    ax2.plot(x2, y2, label=f"$\sigma$ = {sigma} (R = ${wsd.radius}$)")

ax1.legend()
ax1.set_xlabel("x", fontweight="bold")
ax1.set_xticks([-5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5])
ax1.set_ylabel("density", fontweight="bold")

ax2.legend()
ax2.set_xlabel("x", fontweight="bold")
ax2.set_xticks([-10, -8, -6, -4, -2, 0, 2, 4, 6, 8, 10])
ax2.set_ylabel("density", fontweight="bold")

fig.suptitle("Wigner Semicircle probability density function (PDF)", fontweight="bold")
plt.show()

# Similarly, the *cumulative distribution function* (CDF) of the Wigner's Semicircle
# Law can also be analyzed and plotted using **WignerSemicircleDistribution** class.

import numpy as np
import matplotlib.pyplot as plt
from skrmt.ensemble.spectral_law import WignerSemicircleDistribution


x1 = np.linspace(-5, 5, num=2000)
x2 = np.linspace(-10, 10, num=4000)

fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(12,4))

for sigma in [0.5, 1.0, 2.0, 4.0]:
    wsd = WignerSemicircleDistribution(beta=1, center=0.0, sigma=sigma)

    y1 = wsd.cdf(x1)
    y2 = wsd.cdf(x2)

    ax1.plot(x1, y1, label=f"$\sigma$ = {sigma} (R = ${wsd.radius}$)")
    ax2.plot(x2, y2, label=f"$\sigma$ = {sigma} (R = ${wsd.radius}$)")

ax1.legend()
ax1.set_xlabel("x", fontweight="bold")
ax1.set_xticks([-5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5])
ax1.set_ylabel("distribution", fontweight="bold")

ax2.legend()
ax2.set_xlabel("x", fontweight="bold")
ax2.set_xticks([-10, -8, -6, -4, -2, 0, 2, 4, 6, 8, 10])
ax2.set_ylabel("distribution", fontweight="bold")

fig.suptitle("Wigner Semicircle cumulative distribution function (CDF)", fontweight="bold")
plt.show()

##############################################################################
# Tracy-Widom Law
# ---------------
#
# The distribution of the largest eigenvalue of a Wigner matrix converges
# asymptotically (i.e. in the limit of infinite matrix size) to the
# **Tracy-Widom law**. The Tracy-Widom distribution can be defined as
# the limit:
#
#   .. math:: F_2 (s) = \lim_{n \to \infty} \mathbb{P}\left( \sqrt{2} (\lambda_{max} - \sqrt{2 n}) n^{1/6} \le s \right).
#
# The shift :math:`\sqrt{2n}` is used to keep the distributions centered
# at :math:`0`, and the factor :math:`\sqrt{2} n^{1/6}` is used because
# the standard deviation of the distributions scales with :math:`O(n^{-1/6})`.
#
# The probability density function (PDF) and cumulative distribution function (CDF)
# of the Tracy-Widom Law can be computed and graphically represented using the
# **TracyWidomDistribution** class from **scikit-rmt**.

import numpy as np
import matplotlib.pyplot as plt
from skrmt.ensemble.spectral_law import TracyWidomDistribution


x = np.linspace(-5, 2, num=1000)

fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(12,4))

for beta in [1,2,4]:
    twd = TracyWidomDistribution(beta=beta)

    y_pdf = twd.pdf(x)
    y_cdf = twd.cdf(x)

    ax1.plot(x, y_pdf, label=f"$\\beta$ = {beta}")
    ax2.plot(x, y_cdf, label=f"$\\beta$ = {beta}")

ax1.legend()
ax1.set_xlabel("x", fontweight="bold")
ax1.set_ylabel("density", fontweight="bold")
ax1.set_title("Probability density function")

ax2.legend()
ax2.set_xlabel("x", fontweight="bold")
ax2.set_ylabel("distribution", fontweight="bold")
ax2.set_title("Cumulative distribution function")

fig.suptitle("Tracy Widom Law", fontweight="bold")
plt.show()

##############################################################################
# Marchenko-Pastur Law
# --------------------
#
# Another heavily studied type of random matrix are Wishart random matrices.
# The spectrum of the Wishart Ensemble is characterized by the Marchenko-Pastur Law.
# The **Marchenko-Pastur Law** describes the asymptotic behavior of the spectrum
# of a Wishart matrix. Consider the :math:`p \times p` Wishart matrix
# :math:`\mathbf{M} = \sum_{i=1}^{n}\mathbf{x}_i \mathbf{x}_i^\top`, where
# :math:`\mathbf{x}_i \sim N_p(\mathbf{0}, \sigma^2 \mathbf{I}_p)`. In the 
# limit :math:`p,n \to \infty` with :math:`\lambda = p/n \in (0, 1]` fixed,
# the (discrete) distribution of the eigenvalues of :math:`\mathbf{M}`
#
#   .. math:: F^{\mathbf{M}}(x) := \frac{1}{p} \#\{1 \leq i \leq p : \lambda_i(\mathbf{M}) \leq x\},
#
# converges weakly with probability 1 to 
#
#   .. math:: F^{\mathbf{M}}(x)  \underset{\substack{n,p \to \infty \\ p/n \to \lambda}}{\longrightarrow} \int_{-\infty}^{x}f_{\lambda}(t)dt \quad \forall x \in \mathbb{R},
#
# is the Marchenko-Pastur probability density function.
#
# A similar result is obtained if :math:`\lambda > 1`. In this case, the limiting
# distribution has an additional mass probability point in the origin of
# size :math:`1 - \frac{1}{\lambda}`.
#
# Below we show how we can use the class **MarchenkoPasturDistribution** from **scikit-rmt**
# to compute, study and illustrate the PDF of the Marchenko-Pastur Law.

import numpy as np
import matplotlib.pyplot as plt
from skrmt.ensemble.spectral_law import MarchenkoPasturDistribution


x1 = np.linspace(0, 4, num=1000)
x2 = np.linspace(0, 5, num=2000)

fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(12,4))

for ratio in [0.2, 0.4, 0.6, 1.0, 1.4]:
    mpl = MarchenkoPasturDistribution(beta=1, ratio=ratio, sigma=1.0)

    y1 = mpl.pdf(x1)
    y2 = mpl.pdf(x2)

    ax1.plot(x1, y1, label=f"$\lambda$ = {ratio} ")
    ax2.plot(x2, y2, label=f"$\lambda$ = {ratio} ")

ax1.legend()
ax1.set_ylim(0, 1.4)
ax1.set_xlabel("x", fontweight="bold")
ax1.set_ylabel("density", fontweight="bold")

ax2.legend()
ax2.set_ylim(0, 1.4)
ax2.set_xlim(0, 1)
ax2.set_xlabel("x", fontweight="bold")
ax2.set_ylabel("density", fontweight="bold")

fig.suptitle("Marchenko-Pastur probability density function (PDF)", fontweight="bold")
plt.show()

##############################################################################
# Manova spectrum distribution
# ----------------------------
#
# The empirical density of eigenvalues of Manova Ensemble random matrix converges
# almost surely to
#
#   .. math:: f_{M}(x) = (a+b)\frac{\sqrt{(\lambda_+ - x)(x - \lambda_-)}}{2 \pi x (1-x)} I_{[\lambda_- , \lambda_+]} (x),
#
# where :math:`a` and :math:`b` are the matrix parameters, and
#
#   .. math:: \lambda_{\pm} = \left( \sqrt{\frac{a}{a+b}\left( 1 - \frac{1}{a+b} \right)} \pm \sqrt{\frac{1}{a+b}\left(1 - \frac{a}{a+b} \right)} \right)^2, \quad \lambda_{\pm} \in (0,1).
#
# The support of :math:`f_M` is the compact interval :math:`(0,1)`.
#
# The class ManovaSpectrumDistribution can be used to analyze the PDF and CDF of the
# spectrum of the Manova random matrices, as exemplified below.

import numpy as np
import matplotlib.pyplot as plt
from skrmt.ensemble.spectral_law import ManovaSpectrumDistribution

plt.rcParams['figure.dpi'] = 100

x1 = np.linspace(0, 1, num=1000)
x2 = np.linspace(0, 1, num=1000)

fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(12,4))

for a in [1.0, 1.2, 1.4, 1.6]:
    for b in [2.0]:
        msd = ManovaSpectrumDistribution(beta=1, a=a, b=b)

        y1 = msd.pdf(x1)
        y2 = msd.pdf(x2)

        ax1.plot(x1, y1, label=f"$a$ = {a}, $b$ = {b}")
        ax2.plot(x2, y2, label=f"$a$ = {a}, $b$ = {b}")

ax1.legend()
ax1.set_xlabel("x", fontweight="bold")
ax1.set_ylabel("density", fontweight="bold")

ax2.legend()
ax2.set_ylim(0, 4)
ax2.set_xlabel("x", fontweight="bold")
ax2.set_ylabel("density", fontweight="bold")

fig.suptitle("Manova spectrum probability density function (PDF)", fontweight="bold")
plt.show()



"""Wishart Ensemble Module

This module contains the implementation of the Wishart Ensemble, also
known as Laguerre Ensemble. This ensemble of random matrices contains
mainly three sub-ensembles: Wishart Real Ensemble, Wishart Complex Ensemble
and Wishart Quaternion Ensemble.

"""

import numpy as np
import matplotlib.pyplot as plt
from scipy import sparse, special

from ._base_ensemble import _Ensemble
from .tridiagonal_utils import tridiag_eigval_hist


#########################################################################
### Wishart Ensemble = Laguerre Ensemble

class WishartEnsemble(_Ensemble):
    """General Wishart Ensemble class.

    This class contains common attributes and methods for all the
    Wishart ensembles. Wishart Ensembles are divided in:
    - Wishart Real Ensemble (WRE, beta=1): the random matrices of
    this ensemble are formed by multiplying a random real standard
    gaussian matrix of size p times n by its transpose.
    - Wishart Complex Ensemble (WCE, beta=2): the random matrices
    of this ensemble are formed by multiplying a random complex
    standard gaussian matrix of size p times n by its transpose.
    - Wishart Quaternion Ensemble (WQE, beta=4): the random matrices
    of this ensemble are formed by: sampling two random complex
    standard guassian matrices (X and Y), stacking them to create
    matrix A = [X  Y; -conj(Y)  conj(X)]. Finally matrix A is
    multiplied by its transpose in order to generate a matrix of
    the Wishart Quaternion Ensemble.

    Attributes:
        matrix (numpy array): instance of the WishartReal, WishartComplex
            or WishartQuaternion random ensembles. If it is an instance
            of WishartReal or WishartComplex, the random matrix is of
            size p times p. If it is a WishartQuaternion, the random matrix
            is of size 2p times 2p.
        beta (int): descriptive integer of the Wishart ensemble type.
            For Real beta=1, for Complex beta=2, for Quaternion beta=4.
        p (int): number of rows of the guassian matrix that generates
            the matrix of the corresponding ensemble.
        n (int): number of columns of the guassian matrix that generates
            the matrix of the corresponding ensemble.
        use_tridiagonal (bool): if set to True, Gaussian Ensemble
            matrices are sampled in its tridiagonal form, which has the same
            eigenvalues than its standard form. Otherwise, it is sampled using
            its standard form.
        sigma (float): scale (standard deviation) of the random entries of the
            sampled matrix.

    References:
        - Albrecht, J. and Chan, C.P. and Edelman, A.
            "Sturm sequences and random eigenvalue distributions".
            Foundations of Computational Mathematics. 9.4 (2008): 461-483.
        - Dumitriu, I. and Edelman, A. "Matrix Models for Beta Ensembles".
            Journal of Mathematical Physics. 43.11 (2002): 5830-5847.
        - Bar, Z.D. and Silverstain, J.W.
            Spectral Analysis of Large Dimensional Random Matrices.
            2nd edition. Springer. (2010).

    """

    def __init__(self, beta, p, n, use_tridiagonal=False, sigma=1.0):
        """Constructor for WishartEnsemble class.

        Initializes an instance of this class with the given parameters.

        Args:
            beta (int): descriptive integer of the Wishart ensemble type.
                For Real beta=1, for Complex beta=2, for Quaternion beta=4
            p (int): number of rows of the guassian matrix that generates
                the matrix of the corresponding ensemble.
            n (int): number of columns of the guassian matrix that generates
                the matrix of the corresponding ensemble.
            use_tridiagonal (bool, default=False): if set to True, Wishart Ensemble
                matrices are sampled in its tridiagonal form, which has the same
                eigenvalues than its standard form.
            sigma (float, 1.0): scale (standard deviation) of the random entries of the
                sampled matrix.

        """
        if beta not in [1,2,4]:
            raise ValueError(f"Invalid beta: {beta}. Beta value has to be 1, 2 or 4.")

        super().__init__()
        # pylint: disable=invalid-name
        self.p = p
        self.n = n
        self.beta = beta
        self.use_tridiagonal = use_tridiagonal
        self.sigma = sigma
        self._eigvals = None
        self.matrix = self.sample()
        # default eigenvalue normalization constant
        self.eigval_norm_const = 1/self.n

    def set_size(self, p, n, resample_mtx=True):
        # pylint: disable=arguments-differ
        """Setter of matrix size.

        Sets the matrix size. Useful if it has been initialized with a different value.

        Args:
            p (int): number of rows of the guassian matrix that generates
                the matrix of the corresponding ensemble.
            n (int): number of columns of the guassian matrix that generates
                the matrix of the corresponding ensemble.
            resample_mtx (bool, default=True): If set to True, the ensemble matrix is
                resampled with the new dimensions.

        """
        self.p = p
        self.n = n
        if resample_mtx:
            self.matrix = self.sample()

    # pylint: disable=inconsistent-return-statements
    def sample(self):
        """Samples new Wishart Ensemble random matrix.

        The sampling algorithm depends on the specification of
        use_tridiagonal parameter. If use_tridiagonal is set to True,
        a Wishart Ensemble random matrix in its tridiagonal form
        is sampled. Otherwise, it is sampled using the standard
        form.

        Returns:
            numpy array containing new matrix sampled.

        References:
            - Dumitriu, I. and Edelman, A. "Matrix Models for Beta Ensembles".
                Journal of Mathematical Physics. 43.11 (2002): 5830-5847.
        """
        if self.use_tridiagonal:
            if self.p > self.n:  # check reference ("Matrix Models for Beta Ensembles"): page 5, table 1.
                raise ValueError("Error: cannot use tridiagonal form if 'p' (degrees of freedom)"
                                 " is greater than 'n' (sample size).\n"
                                 f"\t Provided n={self.n} and p={self.p}."
                                 " Set `use_tridiagonal=False` or increase sample size (`n`).")
            
            if self.sigma != 1.0:
                raise ValueError("Error: cannot sample tridiagonal random matrix using non-unitary scale"
                                f" (sigma = {self.sigma}).\n"
                                "\t Set `sigma=1.0` (default) or deactivate tridiagonal sampling.")

            return self.sample_tridiagonal()

        if self.beta == 1:
            return self._sample_wre()
        if self.beta == 2:
            return self._sample_wce()
        if self.beta == 4:
            return self._sample_wqe()

    def _sample_wre(self):
        p_size = self.p
        n_size = self.n
        # p by n matrix of random Gaussians
        mtx = np.random.randn(p_size,n_size) * self.sigma
        # symmetrize matrix
        self.matrix = np.matmul(mtx, mtx.transpose())
        # setting array of eigenvalues to None to force re-computing them
        self._eigvals = None
        return self.matrix

    def _sample_wce(self):
        p_size = self.p
        n_size = self.n
        # p by n random complex matrix of random Gaussians
        mtx = np.random.randn(p_size,n_size)*self.sigma + 1j*np.random.randn(p_size,n_size)*self.sigma
        # hermitian matrix
        self.matrix = np.matmul(mtx, mtx.transpose().conj())
        # setting array of eigenvalues to None to force re-computing them
        self._eigvals = None
        return self.matrix

    def _sample_wqe(self):
        p_size = self.p
        n_size = self.n
        # p by n random complex matrix of random Gaussians
        x_mtx = np.random.randn(p_size,n_size)*self.sigma + 1j*np.random.randn(p_size,n_size)*self.sigma
        # p by n random complex matrix of random Gaussians
        y_mtx = np.random.randn(p_size,n_size)*self.sigma + 1j*np.random.randn(p_size,n_size)*self.sigma
        # [X Y; -conj(Y) conj(X)]
        mtx = np.block([
                        [x_mtx              , y_mtx],
                        [-np.conjugate(y_mtx), np.conjugate(x_mtx)]
                    ])
        # hermitian matrix
        self.matrix = np.matmul(mtx, mtx.transpose().conj())
        # setting array of eigenvalues to None to force re-computing them
        self._eigvals = None
        return self.matrix

    def sample_tridiagonal(self):
        '''Samples a Wishart Ensemble random matrix in its tridiagonal form.

        Samples a random matrix of the specified Wishart Ensemble (remember,
        beta=1 is Real, beta=2 is Complex and beta=4 is Quaternion) in its
        tridiagonal form.

        Returns:
            numpy array containing new matrix sampled.

        References:
            - Albrecht, J. and Chan, C.P. and Edelman, A.
                "Sturm sequences and random eigenvalue distributions".
                Foundations of Computational Mathematics. 9.4 (2008): 461-483.
            - Dumitriu, I. and Edelman, A. "Matrix Models for Beta Ensembles".
                Journal of Mathematical Physics. 43.11 (2002): 5830-5847.

        '''
        a_val = self.n*self.beta/2
        # sampling chi-squares
        dfs = np.arange(self.p)
        chisqs_diag = np.array([np.sqrt(np.random.chisquare(2*a_val - self.beta*df)) for df in dfs])
        dfs = np.flip(dfs)
        chisqs_offdiag = np.array([np.sqrt(np.random.chisquare(self.beta*df)) for df in dfs[:-1]])
        # calculating tridiagonal diagonals
        diag = np.array([chisqs_diag[0]**2]+[chisqs_diag[i+1]**2 + \
                         chisqs_offdiag[i]**2 for i in range(self.p-1)])
        offdiag = np.multiply(chisqs_offdiag, chisqs_diag[:-1])
        # inserting diagonals
        diagonals = [offdiag, diag, offdiag]
        mtx = sparse.diags(diagonals, [-1, 0, 1])
        # converting to numpy array
        self.matrix = mtx.toarray()
        # setting array of eigenvalues to None to force re-computing them
        self._eigvals = None
        return self.matrix


    def eigvals(self, normalize=False):
        """Computes the random matrix eigenvalues.

        Calculates the random matrix eigenvalues using numpy standard procedure.
        If the matrix ensemble is symmetric, a faster algorithm is used.

        Returns:
            numpy array with the calculated eigenvalues.

        """
        norm_const = self.eigval_norm_const if normalize else 1.0

        if self._eigvals is not None:
            return norm_const * self._eigvals

        # always storing non-normalized eigenvalues
        self._eigvals = np.linalg.eigvalsh(self.matrix)
        return norm_const * self._eigvals

    def eigval_hist(self, bins, interval=None, density=False, normalize=True, avoid_img=False):
        if self.use_tridiagonal:
            if normalize:
                return tridiag_eigval_hist(
                    self.eigval_norm_const * self.matrix,
                    bins=bins,
                    interval=interval,
                    density=density,
                )
            return tridiag_eigval_hist(self.matrix, bins=bins, interval=interval, density=density)

        return super().eigval_hist(bins, interval=interval, density=density,
                                   normalize=normalize, avoid_img=avoid_img)

    def plot_eigval_hist(self, bins=100, interval=None, density=False, normalize=True, fig_path=None):
        """Computes and plots the histogram of the matrix eigenvalues.

        Calculates and plots the histogram of the current sampled matrix eigenvalues.
        Gaussian (Hermite) ensemble and Wishart (Laguerre) ensemble have improved
        routines to avoid calculating the eigenvalues, so the histogram
        is built using certain techniques to boost efficiency.

        Args:
            bins (int or sequence, default=100): If bins is an integer, it defines the number of
                equal-width bins in the range. If bins is a sequence, it defines the
                bin edges, including the left edge of the first bin and the right
                edge of the last bin; in this case, bins may be unequally spaced.
            interval (tuple, default=None): Delimiters (xmin, xmax) of the histogram.
                The lower and upper range of the bins. Lower and upper outliers are ignored.
            density (bool, default=False): If True, draw and return a probability
                density: each bin will display the bin's raw count divided by the total
                number of counts and the bin width, so that the area under the histogram
                integrates to 1. If set to False, the absolute frequencies of the eigenvalues
                are returned.
            normalize (bool, default=True): Whether to normalize the computed eigenvalues
                by the default normalization constant (see references). Defaults to True, i.e.,
                the eigenvalues are normalized. Normalization makes the eigenvalues to be in the
                same support independently of the sample size.
            fig_path (string, default=None): path to save the created figure. If it is not
                provided, the plot is shown at the end of the routine.

        References:
            - Albrecht, J. and Chan, C.P. and Edelman, A.
                "Sturm sequences and random eigenvalue distributions".
                Foundations of Computational Mathematics. 9.4 (2008): 461-483.
            - Dumitriu, I. and Edelman, A.
                "Matrix Models for Beta Ensembles".
                Journal of Mathematical Physics. 43.11 (2002): 5830-5847.

        """
        if not normalize:
            print("Warning: setting normalize=False may cause normal instability and/or rounding errors.")

        # pylint: disable=too-many-arguments
        if interval is None:
            # calculating constants depending on matrix sizes
            ratio = self.p/self.n
            lambda_plus = self.beta * self.sigma**2 * (1 + np.sqrt(ratio))**2
            lambda_minus = self.beta * self.sigma**2 * (1 - np.sqrt(ratio))**2
            interval = (lambda_minus, lambda_plus)

        if self.use_tridiagonal:
            observed, bins = self.eigval_hist(
                bins=bins, interval=interval, density=density, normalize=normalize
            )

            width = bins[1]-bins[0]
            plt.bar(bins[:-1], observed, width=width, align='edge')
            plt.title("Eigenvalue histogram", fontweight="bold")
            plt.xlabel("x")
            plt.ylabel("density")
            # Saving plot or showing it
            if fig_path:
                plt.savefig(fig_path, dpi=1000)
            else:
                plt.show()

        else:
            super().plot_eigval_hist(
                bins=bins,
                interval=interval,
                density=density,
                normalize=normalize,
                fig_path=fig_path,
            )

    def joint_eigval_pdf(self, eigvals=None):
        '''Computes joint eigenvalue pdf.

        Calculates joint eigenvalue probability density function given an array of
        eigenvalues. If the array of eigenvalues is not provided, the current random
        matrix sample (so its eigenvalues) is used. This function depends on beta,
        i.e., in the sub-Wishart ensemble.

        Args:
            eigvals (np.ndarray, default=None): numpy array with the values (eigenvalues)
                to evaluate the joint pdf in.

        Returns:
            real number. Value of the joint pdf of the eigenvalues.

        References:
            - Dumitriu, I. and Edelman, A. "Matrix Models for Beta Ensembles".
                Journal of Mathematical Physics. 43.11 (2002): 5830-5847.

        '''
        if eigvals is None:
            # calculating eigenvalues
            eigvals = self.eigvals()
        n_eigvals = len(eigvals)

        a_val = self.beta*self.n/2
        p_aux = 1 + self.beta/2*(self.p - 1)
        # calculating Laguerre eigval pdf constant depeding on beta
        const_beta = 2**(-self.p*a_val)
        for j in range(self.p):
            const_beta *= special.gamma(1 + self.beta/2)/ \
                          (special.gamma(1 + self.beta*j/2)*\
                            special.gamma(a_val - self.beta/2*(self.p - j)))
        # calculating first prod
        prod1 = 1
        for j in range(n_eigvals):
            for i in range(j):
                prod1 *= np.abs(eigvals[i] - eigvals[j])**self.beta
        # calculating second prod
        prod2 = np.prod(eigvals**(a_val - p_aux))
        # calculating exponential term
        exp_val = np.exp(-np.sum((eigvals**2)/2))
        # calculating Laguerre eigval pdf
        return const_beta * prod1 * prod2 * exp_val



'''Tridiagonalization Test module

Testing tridiagonalization using Householder reduction
'''

import numpy as np
from numpy.testing import (
    assert_almost_equal,
)

from skrmt.ensemble import householder_reduction


def test_householder_reduction():
    '''Testing householder reduction method for tridiagonalization
    '''
    mtx = np.asarray([[4,2,-2,1], [2,3,2,1], [-2,2,1,0], [1,1,0,2]])
    sol = np.asarray([[4, -3, 0, 0], [-3, 2/3, 5/3, 0], [0, 5/3, 3, 4/3], [0,0,4/3, 7/3]])

    reduction = householder_reduction(mtx)

    assert_almost_equal(reduction, sol, decimal=7)


def test_householder_reduction_eigvals():
    '''Testing householder reduction method keeps eigenvalues after tridiagonalization
    '''
    mtx = np.asarray([[4,2,-2,1], [2,3,2,1], [-2,2,1,0], [1,1,0,2]])
    sol = np.asarray([[4, -3, 0, 0], [-3, 2/3, 5/3, 0], [0, 5/3, 3, 4/3], [0,0,4/3, 7/3]])

    mtx_eigvals = np.linalg.eigvals(mtx)
    sol_eigvals = np.linalg.eigvals(sol)

    reduction = householder_reduction(mtx)
    reduction_eigvals = np.linalg.eigvals(reduction)

    assert_almost_equal(sol_eigvals, reduction_eigvals, decimal=7)
    assert_almost_equal(mtx_eigvals, reduction_eigvals, decimal=7)
